"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import collections.abc
import sys
import typing

import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _BinaryOperator:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _BinaryOperatorEnumTypeWrapper(
    google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_BinaryOperator.ValueType],
    builtins.type,
):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    BINARY_OP_UNSPECIFIED: _BinaryOperator.ValueType  # 0
    BINARY_OP_ADD: _BinaryOperator.ValueType  # 1
    """Arithmetic"""
    BINARY_OP_SUB: _BinaryOperator.ValueType  # 2
    BINARY_OP_MUL: _BinaryOperator.ValueType  # 3
    BINARY_OP_DIV: _BinaryOperator.ValueType  # 4
    BINARY_OP_FLOOR_DIV: _BinaryOperator.ValueType  # 5
    BINARY_OP_MOD: _BinaryOperator.ValueType  # 6
    BINARY_OP_EQ: _BinaryOperator.ValueType  # 10
    """Comparison"""
    BINARY_OP_NE: _BinaryOperator.ValueType  # 11
    BINARY_OP_LT: _BinaryOperator.ValueType  # 12
    BINARY_OP_LE: _BinaryOperator.ValueType  # 13
    BINARY_OP_GT: _BinaryOperator.ValueType  # 14
    BINARY_OP_GE: _BinaryOperator.ValueType  # 15
    BINARY_OP_IN: _BinaryOperator.ValueType  # 16
    BINARY_OP_NOT_IN: _BinaryOperator.ValueType  # 17
    BINARY_OP_AND: _BinaryOperator.ValueType  # 20
    """Logical"""
    BINARY_OP_OR: _BinaryOperator.ValueType  # 21

class BinaryOperator(_BinaryOperator, metaclass=_BinaryOperatorEnumTypeWrapper): ...

BINARY_OP_UNSPECIFIED: BinaryOperator.ValueType  # 0
BINARY_OP_ADD: BinaryOperator.ValueType  # 1
"""Arithmetic"""
BINARY_OP_SUB: BinaryOperator.ValueType  # 2
BINARY_OP_MUL: BinaryOperator.ValueType  # 3
BINARY_OP_DIV: BinaryOperator.ValueType  # 4
BINARY_OP_FLOOR_DIV: BinaryOperator.ValueType  # 5
BINARY_OP_MOD: BinaryOperator.ValueType  # 6
BINARY_OP_EQ: BinaryOperator.ValueType  # 10
"""Comparison"""
BINARY_OP_NE: BinaryOperator.ValueType  # 11
BINARY_OP_LT: BinaryOperator.ValueType  # 12
BINARY_OP_LE: BinaryOperator.ValueType  # 13
BINARY_OP_GT: BinaryOperator.ValueType  # 14
BINARY_OP_GE: BinaryOperator.ValueType  # 15
BINARY_OP_IN: BinaryOperator.ValueType  # 16
BINARY_OP_NOT_IN: BinaryOperator.ValueType  # 17
BINARY_OP_AND: BinaryOperator.ValueType  # 20
"""Logical"""
BINARY_OP_OR: BinaryOperator.ValueType  # 21
Global___BinaryOperator: typing_extensions.TypeAlias = BinaryOperator

class _UnaryOperator:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _UnaryOperatorEnumTypeWrapper(
    google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_UnaryOperator.ValueType],
    builtins.type,
):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    UNARY_OP_UNSPECIFIED: _UnaryOperator.ValueType  # 0
    UNARY_OP_NEG: _UnaryOperator.ValueType  # 1
    UNARY_OP_NOT: _UnaryOperator.ValueType  # 2

class UnaryOperator(_UnaryOperator, metaclass=_UnaryOperatorEnumTypeWrapper): ...

UNARY_OP_UNSPECIFIED: UnaryOperator.ValueType  # 0
UNARY_OP_NEG: UnaryOperator.ValueType  # 1
UNARY_OP_NOT: UnaryOperator.ValueType  # 2
Global___UnaryOperator: typing_extensions.TypeAlias = UnaryOperator

class _GlobalFunction:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _GlobalFunctionEnumTypeWrapper(
    google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_GlobalFunction.ValueType],
    builtins.type,
):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    GLOBAL_FUNCTION_UNSPECIFIED: _GlobalFunction.ValueType  # 0
    GLOBAL_FUNCTION_RANGE: _GlobalFunction.ValueType  # 1
    GLOBAL_FUNCTION_LEN: _GlobalFunction.ValueType  # 2
    GLOBAL_FUNCTION_ENUMERATE: _GlobalFunction.ValueType  # 3
    GLOBAL_FUNCTION_ISEXCEPTION: _GlobalFunction.ValueType  # 4

class GlobalFunction(_GlobalFunction, metaclass=_GlobalFunctionEnumTypeWrapper):
    """Global (built-in) functions supported by the runtime."""

GLOBAL_FUNCTION_UNSPECIFIED: GlobalFunction.ValueType  # 0
GLOBAL_FUNCTION_RANGE: GlobalFunction.ValueType  # 1
GLOBAL_FUNCTION_LEN: GlobalFunction.ValueType  # 2
GLOBAL_FUNCTION_ENUMERATE: GlobalFunction.ValueType  # 3
GLOBAL_FUNCTION_ISEXCEPTION: GlobalFunction.ValueType  # 4
Global___GlobalFunction: typing_extensions.TypeAlias = GlobalFunction

@typing.final
class Program(google.protobuf.message.Message):
    """-----------------------------------------------------------------------------
    Top-Level Structure
    -----------------------------------------------------------------------------

    A complete Waymark program (one or more function definitions)
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FUNCTIONS_FIELD_NUMBER: builtins.int
    @property
    def functions(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        Global___FunctionDef
    ]: ...
    def __init__(
        self,
        *,
        functions: collections.abc.Iterable[Global___FunctionDef] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["functions", b"functions"]) -> None: ...

Global___Program: typing_extensions.TypeAlias = Program

@typing.final
class FunctionDef(google.protobuf.message.Message):
    """Function definition with explicit input/output declarations"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    IO_FIELD_NUMBER: builtins.int
    BODY_FIELD_NUMBER: builtins.int
    SPAN_FIELD_NUMBER: builtins.int
    name: builtins.str
    @property
    def io(self) -> Global___IoDecl: ...
    @property
    def body(self) -> Global___Block: ...
    @property
    def span(self) -> Global___Span: ...
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        io: Global___IoDecl | None = ...,
        body: Global___Block | None = ...,
        span: Global___Span | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing.Literal["body", b"body", "io", b"io", "span", b"span"]
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal["body", b"body", "io", b"io", "name", b"name", "span", b"span"],
    ) -> None: ...

Global___FunctionDef: typing_extensions.TypeAlias = FunctionDef

@typing.final
class IoDecl(google.protobuf.message.Message):
    """Input/output declaration: fn foo(input: [a, b], output: [c])"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    INPUTS_FIELD_NUMBER: builtins.int
    OUTPUTS_FIELD_NUMBER: builtins.int
    SPAN_FIELD_NUMBER: builtins.int
    @property
    def inputs(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    @property
    def outputs(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    @property
    def span(self) -> Global___Span: ...
    def __init__(
        self,
        *,
        inputs: collections.abc.Iterable[builtins.str] | None = ...,
        outputs: collections.abc.Iterable[builtins.str] | None = ...,
        span: Global___Span | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["span", b"span"]) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal["inputs", b"inputs", "outputs", b"outputs", "span", b"span"],
    ) -> None: ...

Global___IoDecl: typing_extensions.TypeAlias = IoDecl

@typing.final
class Block(google.protobuf.message.Message):
    """A block of statements (function body - no call restrictions)"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STATEMENTS_FIELD_NUMBER: builtins.int
    SPAN_FIELD_NUMBER: builtins.int
    @property
    def statements(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        Global___Statement
    ]: ...
    @property
    def span(self) -> Global___Span: ...
    def __init__(
        self,
        *,
        statements: collections.abc.Iterable[Global___Statement] | None = ...,
        span: Global___Span | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["span", b"span"]) -> builtins.bool: ...
    def ClearField(
        self, field_name: typing.Literal["span", b"span", "statements", b"statements"]
    ) -> None: ...

Global___Block: typing_extensions.TypeAlias = Block

@typing.final
class Statement(google.protobuf.message.Message):
    """-----------------------------------------------------------------------------
    Statements
    -----------------------------------------------------------------------------
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ASSIGNMENT_FIELD_NUMBER: builtins.int
    ACTION_CALL_FIELD_NUMBER: builtins.int
    SPREAD_ACTION_FIELD_NUMBER: builtins.int
    PARALLEL_BLOCK_FIELD_NUMBER: builtins.int
    FOR_LOOP_FIELD_NUMBER: builtins.int
    CONDITIONAL_FIELD_NUMBER: builtins.int
    TRY_EXCEPT_FIELD_NUMBER: builtins.int
    RETURN_STMT_FIELD_NUMBER: builtins.int
    EXPR_STMT_FIELD_NUMBER: builtins.int
    BREAK_STMT_FIELD_NUMBER: builtins.int
    CONTINUE_STMT_FIELD_NUMBER: builtins.int
    WHILE_LOOP_FIELD_NUMBER: builtins.int
    SLEEP_STMT_FIELD_NUMBER: builtins.int
    SPAN_FIELD_NUMBER: builtins.int
    @property
    def assignment(self) -> Global___Assignment: ...
    @property
    def action_call(self) -> Global___ActionCall: ...
    @property
    def spread_action(self) -> Global___SpreadAction: ...
    @property
    def parallel_block(self) -> Global___ParallelBlock: ...
    @property
    def for_loop(self) -> Global___ForLoop: ...
    @property
    def conditional(self) -> Global___Conditional: ...
    @property
    def try_except(self) -> Global___TryExcept: ...
    @property
    def return_stmt(self) -> Global___ReturnStmt: ...
    @property
    def expr_stmt(self) -> Global___ExprStmt: ...
    @property
    def break_stmt(self) -> Global___BreakStmt: ...
    @property
    def continue_stmt(self) -> Global___ContinueStmt: ...
    @property
    def while_loop(self) -> Global___WhileLoop: ...
    @property
    def sleep_stmt(self) -> Global___SleepStmt: ...
    @property
    def span(self) -> Global___Span: ...
    def __init__(
        self,
        *,
        assignment: Global___Assignment | None = ...,
        action_call: Global___ActionCall | None = ...,
        spread_action: Global___SpreadAction | None = ...,
        parallel_block: Global___ParallelBlock | None = ...,
        for_loop: Global___ForLoop | None = ...,
        conditional: Global___Conditional | None = ...,
        try_except: Global___TryExcept | None = ...,
        return_stmt: Global___ReturnStmt | None = ...,
        expr_stmt: Global___ExprStmt | None = ...,
        break_stmt: Global___BreakStmt | None = ...,
        continue_stmt: Global___ContinueStmt | None = ...,
        while_loop: Global___WhileLoop | None = ...,
        sleep_stmt: Global___SleepStmt | None = ...,
        span: Global___Span | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing.Literal[
            "action_call",
            b"action_call",
            "assignment",
            b"assignment",
            "break_stmt",
            b"break_stmt",
            "conditional",
            b"conditional",
            "continue_stmt",
            b"continue_stmt",
            "expr_stmt",
            b"expr_stmt",
            "for_loop",
            b"for_loop",
            "kind",
            b"kind",
            "parallel_block",
            b"parallel_block",
            "return_stmt",
            b"return_stmt",
            "sleep_stmt",
            b"sleep_stmt",
            "span",
            b"span",
            "spread_action",
            b"spread_action",
            "try_except",
            b"try_except",
            "while_loop",
            b"while_loop",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "action_call",
            b"action_call",
            "assignment",
            b"assignment",
            "break_stmt",
            b"break_stmt",
            "conditional",
            b"conditional",
            "continue_stmt",
            b"continue_stmt",
            "expr_stmt",
            b"expr_stmt",
            "for_loop",
            b"for_loop",
            "kind",
            b"kind",
            "parallel_block",
            b"parallel_block",
            "return_stmt",
            b"return_stmt",
            "sleep_stmt",
            b"sleep_stmt",
            "span",
            b"span",
            "spread_action",
            b"spread_action",
            "try_except",
            b"try_except",
            "while_loop",
            b"while_loop",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing.Literal["kind", b"kind"]
    ) -> (
        typing.Literal[
            "assignment",
            "action_call",
            "spread_action",
            "parallel_block",
            "for_loop",
            "conditional",
            "try_except",
            "return_stmt",
            "expr_stmt",
            "break_stmt",
            "continue_stmt",
            "while_loop",
            "sleep_stmt",
        ]
        | None
    ): ...

Global___Statement: typing_extensions.TypeAlias = Statement

@typing.final
class BreakStmt(google.protobuf.message.Message):
    """Break statement (exits innermost loop)"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

Global___BreakStmt: typing_extensions.TypeAlias = BreakStmt

@typing.final
class ContinueStmt(google.protobuf.message.Message):
    """Continue statement (skips to next iteration of innermost loop)"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

Global___ContinueStmt: typing_extensions.TypeAlias = ContinueStmt

@typing.final
class Assignment(google.protobuf.message.Message):
    """Variable assignment: x = expr"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TARGETS_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    @property
    def targets(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """Single variable or tuple unpacking (a, b = expr)"""

    @property
    def value(self) -> Global___Expr: ...
    def __init__(
        self,
        *,
        targets: collections.abc.Iterable[builtins.str] | None = ...,
        value: Global___Expr | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
    def ClearField(
        self, field_name: typing.Literal["targets", b"targets", "value", b"value"]
    ) -> None: ...

Global___Assignment: typing_extensions.TypeAlias = Assignment

@typing.final
class ActionCall(google.protobuf.message.Message):
    """Action call: @action(kwargs) [retry_policy] [timeout]
    Can be used as:
    - Statement (side effect only): @notify(user=x)
    - Expression in Assignment (with binding): result = @fetch(id=x)
    - Expression in Assignment (with unpacking): a, b = @get_pair()
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ACTION_NAME_FIELD_NUMBER: builtins.int
    KWARGS_FIELD_NUMBER: builtins.int
    POLICIES_FIELD_NUMBER: builtins.int
    MODULE_NAME_FIELD_NUMBER: builtins.int
    action_name: builtins.str
    module_name: builtins.str
    """Python module containing the action"""
    @property
    def kwargs(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[Global___Kwarg]: ...
    @property
    def policies(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        Global___PolicyBracket
    ]: ...
    def __init__(
        self,
        *,
        action_name: builtins.str = ...,
        kwargs: collections.abc.Iterable[Global___Kwarg] | None = ...,
        policies: collections.abc.Iterable[Global___PolicyBracket] | None = ...,
        module_name: builtins.str | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing.Literal["_module_name", b"_module_name", "module_name", b"module_name"],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "_module_name",
            b"_module_name",
            "action_name",
            b"action_name",
            "kwargs",
            b"kwargs",
            "module_name",
            b"module_name",
            "policies",
            b"policies",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing.Literal["_module_name", b"_module_name"]
    ) -> typing.Literal["module_name"] | None: ...

Global___ActionCall: typing_extensions.TypeAlias = ActionCall

@typing.final
class SpreadAction(google.protobuf.message.Message):
    """Spread action statement (side effect only, no result capture)
    For result capture, use Assignment with SpreadExpr
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    COLLECTION_FIELD_NUMBER: builtins.int
    LOOP_VAR_FIELD_NUMBER: builtins.int
    ACTION_FIELD_NUMBER: builtins.int
    loop_var: builtins.str
    """Variable name for each item"""
    @property
    def collection(self) -> Global___Expr:
        """The collection to iterate"""

    @property
    def action(self) -> Global___ActionCall:
        """The action to call per item"""

    def __init__(
        self,
        *,
        collection: Global___Expr | None = ...,
        loop_var: builtins.str = ...,
        action: Global___ActionCall | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing.Literal["action", b"action", "collection", b"collection"]
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "action", b"action", "collection", b"collection", "loop_var", b"loop_var"
        ],
    ) -> None: ...

Global___SpreadAction: typing_extensions.TypeAlias = SpreadAction

@typing.final
class ParallelBlock(google.protobuf.message.Message):
    """Parallel block statement (side effect only, no result capture)
    For result capture, use Assignment with ParallelExpr
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CALLS_FIELD_NUMBER: builtins.int
    @property
    def calls(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[Global___Call]:
        """Actions or function calls to run in parallel"""

    def __init__(
        self,
        *,
        calls: collections.abc.Iterable[Global___Call] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["calls", b"calls"]) -> None: ...

Global___ParallelBlock: typing_extensions.TypeAlias = ParallelBlock

@typing.final
class Call(google.protobuf.message.Message):
    """A call that can be either action or function (for parallel blocks)"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ACTION_FIELD_NUMBER: builtins.int
    FUNCTION_FIELD_NUMBER: builtins.int
    @property
    def action(self) -> Global___ActionCall: ...
    @property
    def function(self) -> Global___FunctionCall: ...
    def __init__(
        self,
        *,
        action: Global___ActionCall | None = ...,
        function: Global___FunctionCall | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing.Literal["action", b"action", "function", b"function", "kind", b"kind"],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal["action", b"action", "function", b"function", "kind", b"kind"],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing.Literal["kind", b"kind"]
    ) -> typing.Literal["action", "function"] | None: ...

Global___Call: typing_extensions.TypeAlias = Call

@typing.final
class ForLoop(google.protobuf.message.Message):
    """For loop: for item in items: ..."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LOOP_VARS_FIELD_NUMBER: builtins.int
    ITERABLE_FIELD_NUMBER: builtins.int
    BLOCK_BODY_FIELD_NUMBER: builtins.int
    @property
    def loop_vars(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """Loop variable(s) for unpacking"""

    @property
    def iterable(self) -> Global___Expr: ...
    @property
    def block_body(self) -> Global___Block: ...
    def __init__(
        self,
        *,
        loop_vars: collections.abc.Iterable[builtins.str] | None = ...,
        iterable: Global___Expr | None = ...,
        block_body: Global___Block | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing.Literal["block_body", b"block_body", "iterable", b"iterable"]
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "block_body", b"block_body", "iterable", b"iterable", "loop_vars", b"loop_vars"
        ],
    ) -> None: ...

Global___ForLoop: typing_extensions.TypeAlias = ForLoop

@typing.final
class WhileLoop(google.protobuf.message.Message):
    """While loop: while condition: ..."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CONDITION_FIELD_NUMBER: builtins.int
    BLOCK_BODY_FIELD_NUMBER: builtins.int
    @property
    def condition(self) -> Global___Expr: ...
    @property
    def block_body(self) -> Global___Block: ...
    def __init__(
        self,
        *,
        condition: Global___Expr | None = ...,
        block_body: Global___Block | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing.Literal["block_body", b"block_body", "condition", b"condition"]
    ) -> builtins.bool: ...
    def ClearField(
        self, field_name: typing.Literal["block_body", b"block_body", "condition", b"condition"]
    ) -> None: ...

Global___WhileLoop: typing_extensions.TypeAlias = WhileLoop

@typing.final
class Conditional(google.protobuf.message.Message):
    """Conditional: if/elif/else"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    IF_BRANCH_FIELD_NUMBER: builtins.int
    ELIF_BRANCHES_FIELD_NUMBER: builtins.int
    ELSE_BRANCH_FIELD_NUMBER: builtins.int
    @property
    def if_branch(self) -> Global___IfBranch: ...
    @property
    def elif_branches(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        Global___ElifBranch
    ]: ...
    @property
    def else_branch(self) -> Global___ElseBranch: ...
    def __init__(
        self,
        *,
        if_branch: Global___IfBranch | None = ...,
        elif_branches: collections.abc.Iterable[Global___ElifBranch] | None = ...,
        else_branch: Global___ElseBranch | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing.Literal[
            "_else_branch",
            b"_else_branch",
            "else_branch",
            b"else_branch",
            "if_branch",
            b"if_branch",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "_else_branch",
            b"_else_branch",
            "elif_branches",
            b"elif_branches",
            "else_branch",
            b"else_branch",
            "if_branch",
            b"if_branch",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing.Literal["_else_branch", b"_else_branch"]
    ) -> typing.Literal["else_branch"] | None: ...

Global___Conditional: typing_extensions.TypeAlias = Conditional

@typing.final
class IfBranch(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CONDITION_FIELD_NUMBER: builtins.int
    SPAN_FIELD_NUMBER: builtins.int
    BLOCK_BODY_FIELD_NUMBER: builtins.int
    @property
    def condition(self) -> Global___Expr: ...
    @property
    def span(self) -> Global___Span: ...
    @property
    def block_body(self) -> Global___Block: ...
    def __init__(
        self,
        *,
        condition: Global___Expr | None = ...,
        span: Global___Span | None = ...,
        block_body: Global___Block | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing.Literal[
            "block_body", b"block_body", "condition", b"condition", "span", b"span"
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "block_body", b"block_body", "condition", b"condition", "span", b"span"
        ],
    ) -> None: ...

Global___IfBranch: typing_extensions.TypeAlias = IfBranch

@typing.final
class ElifBranch(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CONDITION_FIELD_NUMBER: builtins.int
    SPAN_FIELD_NUMBER: builtins.int
    BLOCK_BODY_FIELD_NUMBER: builtins.int
    @property
    def condition(self) -> Global___Expr: ...
    @property
    def span(self) -> Global___Span: ...
    @property
    def block_body(self) -> Global___Block: ...
    def __init__(
        self,
        *,
        condition: Global___Expr | None = ...,
        span: Global___Span | None = ...,
        block_body: Global___Block | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing.Literal[
            "block_body", b"block_body", "condition", b"condition", "span", b"span"
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "block_body", b"block_body", "condition", b"condition", "span", b"span"
        ],
    ) -> None: ...

Global___ElifBranch: typing_extensions.TypeAlias = ElifBranch

@typing.final
class ElseBranch(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SPAN_FIELD_NUMBER: builtins.int
    BLOCK_BODY_FIELD_NUMBER: builtins.int
    @property
    def span(self) -> Global___Span: ...
    @property
    def block_body(self) -> Global___Block: ...
    def __init__(
        self,
        *,
        span: Global___Span | None = ...,
        block_body: Global___Block | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing.Literal["block_body", b"block_body", "span", b"span"]
    ) -> builtins.bool: ...
    def ClearField(
        self, field_name: typing.Literal["block_body", b"block_body", "span", b"span"]
    ) -> None: ...

Global___ElseBranch: typing_extensions.TypeAlias = ElseBranch

@typing.final
class TryExcept(google.protobuf.message.Message):
    """Try/except block"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    HANDLERS_FIELD_NUMBER: builtins.int
    TRY_BLOCK_FIELD_NUMBER: builtins.int
    @property
    def handlers(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        Global___ExceptHandler
    ]: ...
    @property
    def try_block(self) -> Global___Block: ...
    def __init__(
        self,
        *,
        handlers: collections.abc.Iterable[Global___ExceptHandler] | None = ...,
        try_block: Global___Block | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["try_block", b"try_block"]) -> builtins.bool: ...
    def ClearField(
        self, field_name: typing.Literal["handlers", b"handlers", "try_block", b"try_block"]
    ) -> None: ...

Global___TryExcept: typing_extensions.TypeAlias = TryExcept

@typing.final
class ExceptHandler(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    EXCEPTION_TYPES_FIELD_NUMBER: builtins.int
    SPAN_FIELD_NUMBER: builtins.int
    BLOCK_BODY_FIELD_NUMBER: builtins.int
    EXCEPTION_VAR_FIELD_NUMBER: builtins.int
    exception_var: builtins.str
    """Variable name for "except ... as var" """
    @property
    def exception_types(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """Empty = catch all"""

    @property
    def span(self) -> Global___Span: ...
    @property
    def block_body(self) -> Global___Block: ...
    def __init__(
        self,
        *,
        exception_types: collections.abc.Iterable[builtins.str] | None = ...,
        span: Global___Span | None = ...,
        block_body: Global___Block | None = ...,
        exception_var: builtins.str | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing.Literal[
            "_exception_var",
            b"_exception_var",
            "block_body",
            b"block_body",
            "exception_var",
            b"exception_var",
            "span",
            b"span",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "_exception_var",
            b"_exception_var",
            "block_body",
            b"block_body",
            "exception_types",
            b"exception_types",
            "exception_var",
            b"exception_var",
            "span",
            b"span",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing.Literal["_exception_var", b"_exception_var"]
    ) -> typing.Literal["exception_var"] | None: ...

Global___ExceptHandler: typing_extensions.TypeAlias = ExceptHandler

@typing.final
class ReturnStmt(google.protobuf.message.Message):
    """Return statement"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUE_FIELD_NUMBER: builtins.int
    @property
    def value(self) -> Global___Expr: ...
    def __init__(
        self,
        *,
        value: Global___Expr | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing.Literal["_value", b"_value", "value", b"value"]
    ) -> builtins.bool: ...
    def ClearField(
        self, field_name: typing.Literal["_value", b"_value", "value", b"value"]
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing.Literal["_value", b"_value"]
    ) -> typing.Literal["value"] | None: ...

Global___ReturnStmt: typing_extensions.TypeAlias = ReturnStmt

@typing.final
class ExprStmt(google.protobuf.message.Message):
    """Expression as statement"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    EXPR_FIELD_NUMBER: builtins.int
    @property
    def expr(self) -> Global___Expr: ...
    def __init__(
        self,
        *,
        expr: Global___Expr | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["expr", b"expr"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["expr", b"expr"]) -> None: ...

Global___ExprStmt: typing_extensions.TypeAlias = ExprStmt

@typing.final
class SleepStmt(google.protobuf.message.Message):
    """Sleep statement (durable sleep handled by scheduler)"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DURATION_FIELD_NUMBER: builtins.int
    @property
    def duration(self) -> Global___Expr: ...
    def __init__(
        self,
        *,
        duration: Global___Expr | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing.Literal["_duration", b"_duration", "duration", b"duration"]
    ) -> builtins.bool: ...
    def ClearField(
        self, field_name: typing.Literal["_duration", b"_duration", "duration", b"duration"]
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing.Literal["_duration", b"_duration"]
    ) -> typing.Literal["duration"] | None: ...

Global___SleepStmt: typing_extensions.TypeAlias = SleepStmt

@typing.final
class Expr(google.protobuf.message.Message):
    """-----------------------------------------------------------------------------
    Expressions
    -----------------------------------------------------------------------------
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LITERAL_FIELD_NUMBER: builtins.int
    VARIABLE_FIELD_NUMBER: builtins.int
    BINARY_OP_FIELD_NUMBER: builtins.int
    UNARY_OP_FIELD_NUMBER: builtins.int
    LIST_FIELD_NUMBER: builtins.int
    DICT_FIELD_NUMBER: builtins.int
    INDEX_FIELD_NUMBER: builtins.int
    DOT_FIELD_NUMBER: builtins.int
    FUNCTION_CALL_FIELD_NUMBER: builtins.int
    ACTION_CALL_FIELD_NUMBER: builtins.int
    PARALLEL_EXPR_FIELD_NUMBER: builtins.int
    SPREAD_EXPR_FIELD_NUMBER: builtins.int
    SPAN_FIELD_NUMBER: builtins.int
    @property
    def literal(self) -> Global___Literal: ...
    @property
    def variable(self) -> Global___Variable: ...
    @property
    def binary_op(self) -> Global___BinaryOp: ...
    @property
    def unary_op(self) -> Global___UnaryOp: ...
    @property
    def list(self) -> Global___ListExpr: ...
    @property
    def dict(self) -> Global___DictExpr: ...
    @property
    def index(self) -> Global___IndexAccess: ...
    @property
    def dot(self) -> Global___DotAccess: ...
    @property
    def function_call(self) -> Global___FunctionCall: ...
    @property
    def action_call(self) -> Global___ActionCall: ...
    @property
    def parallel_expr(self) -> Global___ParallelExpr:
        """Parallel block as expression"""

    @property
    def spread_expr(self) -> Global___SpreadExpr:
        """Spread action as expression"""

    @property
    def span(self) -> Global___Span: ...
    def __init__(
        self,
        *,
        literal: Global___Literal | None = ...,
        variable: Global___Variable | None = ...,
        binary_op: Global___BinaryOp | None = ...,
        unary_op: Global___UnaryOp | None = ...,
        list: Global___ListExpr | None = ...,
        dict: Global___DictExpr | None = ...,
        index: Global___IndexAccess | None = ...,
        dot: Global___DotAccess | None = ...,
        function_call: Global___FunctionCall | None = ...,
        action_call: Global___ActionCall | None = ...,
        parallel_expr: Global___ParallelExpr | None = ...,
        spread_expr: Global___SpreadExpr | None = ...,
        span: Global___Span | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing.Literal[
            "action_call",
            b"action_call",
            "binary_op",
            b"binary_op",
            "dict",
            b"dict",
            "dot",
            b"dot",
            "function_call",
            b"function_call",
            "index",
            b"index",
            "kind",
            b"kind",
            "list",
            b"list",
            "literal",
            b"literal",
            "parallel_expr",
            b"parallel_expr",
            "span",
            b"span",
            "spread_expr",
            b"spread_expr",
            "unary_op",
            b"unary_op",
            "variable",
            b"variable",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "action_call",
            b"action_call",
            "binary_op",
            b"binary_op",
            "dict",
            b"dict",
            "dot",
            b"dot",
            "function_call",
            b"function_call",
            "index",
            b"index",
            "kind",
            b"kind",
            "list",
            b"list",
            "literal",
            b"literal",
            "parallel_expr",
            b"parallel_expr",
            "span",
            b"span",
            "spread_expr",
            b"spread_expr",
            "unary_op",
            b"unary_op",
            "variable",
            b"variable",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing.Literal["kind", b"kind"]
    ) -> (
        typing.Literal[
            "literal",
            "variable",
            "binary_op",
            "unary_op",
            "list",
            "dict",
            "index",
            "dot",
            "function_call",
            "action_call",
            "parallel_expr",
            "spread_expr",
        ]
        | None
    ): ...

Global___Expr: typing_extensions.TypeAlias = Expr

@typing.final
class Literal(google.protobuf.message.Message):
    """Literals: 42, 3.14, "hello", True, False"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    INT_VALUE_FIELD_NUMBER: builtins.int
    FLOAT_VALUE_FIELD_NUMBER: builtins.int
    STRING_VALUE_FIELD_NUMBER: builtins.int
    BOOL_VALUE_FIELD_NUMBER: builtins.int
    IS_NONE_FIELD_NUMBER: builtins.int
    int_value: builtins.int
    float_value: builtins.float
    string_value: builtins.str
    bool_value: builtins.bool
    is_none: builtins.bool
    """true if None"""
    def __init__(
        self,
        *,
        int_value: builtins.int = ...,
        float_value: builtins.float = ...,
        string_value: builtins.str = ...,
        bool_value: builtins.bool = ...,
        is_none: builtins.bool = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing.Literal[
            "bool_value",
            b"bool_value",
            "float_value",
            b"float_value",
            "int_value",
            b"int_value",
            "is_none",
            b"is_none",
            "string_value",
            b"string_value",
            "value",
            b"value",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "bool_value",
            b"bool_value",
            "float_value",
            b"float_value",
            "int_value",
            b"int_value",
            "is_none",
            b"is_none",
            "string_value",
            b"string_value",
            "value",
            b"value",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing.Literal["value", b"value"]
    ) -> (
        typing.Literal["int_value", "float_value", "string_value", "bool_value", "is_none"] | None
    ): ...

Global___Literal: typing_extensions.TypeAlias = Literal

@typing.final
class Variable(google.protobuf.message.Message):
    """Variable reference: foo"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    name: builtins.str
    def __init__(
        self,
        *,
        name: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["name", b"name"]) -> None: ...

Global___Variable: typing_extensions.TypeAlias = Variable

@typing.final
class BinaryOp(google.protobuf.message.Message):
    """Binary operation: a + b, x and y, etc."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LEFT_FIELD_NUMBER: builtins.int
    OP_FIELD_NUMBER: builtins.int
    RIGHT_FIELD_NUMBER: builtins.int
    op: Global___BinaryOperator.ValueType
    @property
    def left(self) -> Global___Expr: ...
    @property
    def right(self) -> Global___Expr: ...
    def __init__(
        self,
        *,
        left: Global___Expr | None = ...,
        op: Global___BinaryOperator.ValueType = ...,
        right: Global___Expr | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing.Literal["left", b"left", "right", b"right"]
    ) -> builtins.bool: ...
    def ClearField(
        self, field_name: typing.Literal["left", b"left", "op", b"op", "right", b"right"]
    ) -> None: ...

Global___BinaryOp: typing_extensions.TypeAlias = BinaryOp

@typing.final
class UnaryOp(google.protobuf.message.Message):
    """Unary operation: -x, not x"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    OP_FIELD_NUMBER: builtins.int
    OPERAND_FIELD_NUMBER: builtins.int
    op: Global___UnaryOperator.ValueType
    @property
    def operand(self) -> Global___Expr: ...
    def __init__(
        self,
        *,
        op: Global___UnaryOperator.ValueType = ...,
        operand: Global___Expr | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["operand", b"operand"]) -> builtins.bool: ...
    def ClearField(
        self, field_name: typing.Literal["op", b"op", "operand", b"operand"]
    ) -> None: ...

Global___UnaryOp: typing_extensions.TypeAlias = UnaryOp

@typing.final
class ListExpr(google.protobuf.message.Message):
    """List literal: [1, 2, 3]"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ELEMENTS_FIELD_NUMBER: builtins.int
    @property
    def elements(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[Global___Expr]: ...
    def __init__(
        self,
        *,
        elements: collections.abc.Iterable[Global___Expr] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["elements", b"elements"]) -> None: ...

Global___ListExpr: typing_extensions.TypeAlias = ListExpr

@typing.final
class DictExpr(google.protobuf.message.Message):
    """Dict literal: {"a": 1, "b": 2}"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ENTRIES_FIELD_NUMBER: builtins.int
    @property
    def entries(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        Global___DictEntry
    ]: ...
    def __init__(
        self,
        *,
        entries: collections.abc.Iterable[Global___DictEntry] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["entries", b"entries"]) -> None: ...

Global___DictExpr: typing_extensions.TypeAlias = DictExpr

@typing.final
class DictEntry(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KEY_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    @property
    def key(self) -> Global___Expr: ...
    @property
    def value(self) -> Global___Expr: ...
    def __init__(
        self,
        *,
        key: Global___Expr | None = ...,
        value: Global___Expr | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing.Literal["key", b"key", "value", b"value"]
    ) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

Global___DictEntry: typing_extensions.TypeAlias = DictEntry

@typing.final
class IndexAccess(google.protobuf.message.Message):
    """Index access: items[0]"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    OBJECT_FIELD_NUMBER: builtins.int
    INDEX_FIELD_NUMBER: builtins.int
    @property
    def object(self) -> Global___Expr: ...
    @property
    def index(self) -> Global___Expr: ...
    def __init__(
        self,
        *,
        object: Global___Expr | None = ...,
        index: Global___Expr | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing.Literal["index", b"index", "object", b"object"]
    ) -> builtins.bool: ...
    def ClearField(
        self, field_name: typing.Literal["index", b"index", "object", b"object"]
    ) -> None: ...

Global___IndexAccess: typing_extensions.TypeAlias = IndexAccess

@typing.final
class DotAccess(google.protobuf.message.Message):
    """Dot access: obj.attr"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    OBJECT_FIELD_NUMBER: builtins.int
    ATTRIBUTE_FIELD_NUMBER: builtins.int
    attribute: builtins.str
    @property
    def object(self) -> Global___Expr: ...
    def __init__(
        self,
        *,
        object: Global___Expr | None = ...,
        attribute: builtins.str = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["object", b"object"]) -> builtins.bool: ...
    def ClearField(
        self, field_name: typing.Literal["attribute", b"attribute", "object", b"object"]
    ) -> None: ...

Global___DotAccess: typing_extensions.TypeAlias = DotAccess

@typing.final
class FunctionCall(google.protobuf.message.Message):
    """Function call: range(10), len(items)
    Built-in functions can use positional args, user functions use kwargs
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    ARGS_FIELD_NUMBER: builtins.int
    KWARGS_FIELD_NUMBER: builtins.int
    GLOBAL_FUNCTION_FIELD_NUMBER: builtins.int
    name: builtins.str
    global_function: Global___GlobalFunction.ValueType
    """Global function identifier (if applicable)"""
    @property
    def args(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[Global___Expr]:
        """Positional arguments (for built-ins)"""

    @property
    def kwargs(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[Global___Kwarg]:
        """Keyword arguments"""

    def __init__(
        self,
        *,
        name: builtins.str = ...,
        args: collections.abc.Iterable[Global___Expr] | None = ...,
        kwargs: collections.abc.Iterable[Global___Kwarg] | None = ...,
        global_function: Global___GlobalFunction.ValueType = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "args",
            b"args",
            "global_function",
            b"global_function",
            "kwargs",
            b"kwargs",
            "name",
            b"name",
        ],
    ) -> None: ...

Global___FunctionCall: typing_extensions.TypeAlias = FunctionCall

@typing.final
class Kwarg(google.protobuf.message.Message):
    """Keyword argument: name=value"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    name: builtins.str
    @property
    def value(self) -> Global___Expr: ...
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        value: Global___Expr | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
    def ClearField(
        self, field_name: typing.Literal["name", b"name", "value", b"value"]
    ) -> None: ...

Global___Kwarg: typing_extensions.TypeAlias = Kwarg

@typing.final
class ParallelExpr(google.protobuf.message.Message):
    """Parallel expression: concurrent execution of multiple calls, returns list
    Used in assignments: a, b = parallel: @x() @y()
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CALLS_FIELD_NUMBER: builtins.int
    @property
    def calls(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[Global___Call]:
        """Actions or function calls to run in parallel"""

    def __init__(
        self,
        *,
        calls: collections.abc.Iterable[Global___Call] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["calls", b"calls"]) -> None: ...

Global___ParallelExpr: typing_extensions.TypeAlias = ParallelExpr

@typing.final
class SpreadExpr(google.protobuf.message.Message):
    """Spread expression: parallel execution over a collection, returns list
    Used in assignments: results = spread items:item -> @action(item=item)
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    COLLECTION_FIELD_NUMBER: builtins.int
    LOOP_VAR_FIELD_NUMBER: builtins.int
    ACTION_FIELD_NUMBER: builtins.int
    loop_var: builtins.str
    """Variable name for each item"""
    @property
    def collection(self) -> Global___Expr:
        """The collection to iterate"""

    @property
    def action(self) -> Global___ActionCall:
        """The action to call per item"""

    def __init__(
        self,
        *,
        collection: Global___Expr | None = ...,
        loop_var: builtins.str = ...,
        action: Global___ActionCall | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing.Literal["action", b"action", "collection", b"collection"]
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "action", b"action", "collection", b"collection", "loop_var", b"loop_var"
        ],
    ) -> None: ...

Global___SpreadExpr: typing_extensions.TypeAlias = SpreadExpr

@typing.final
class PolicyBracket(google.protobuf.message.Message):
    """-----------------------------------------------------------------------------
    Policies (for actions)
    -----------------------------------------------------------------------------

    Policy bracket on an action call
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RETRY_FIELD_NUMBER: builtins.int
    TIMEOUT_FIELD_NUMBER: builtins.int
    @property
    def retry(self) -> Global___RetryPolicy: ...
    @property
    def timeout(self) -> Global___TimeoutPolicy: ...
    def __init__(
        self,
        *,
        retry: Global___RetryPolicy | None = ...,
        timeout: Global___TimeoutPolicy | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing.Literal["kind", b"kind", "retry", b"retry", "timeout", b"timeout"]
    ) -> builtins.bool: ...
    def ClearField(
        self, field_name: typing.Literal["kind", b"kind", "retry", b"retry", "timeout", b"timeout"]
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing.Literal["kind", b"kind"]
    ) -> typing.Literal["retry", "timeout"] | None: ...

Global___PolicyBracket: typing_extensions.TypeAlias = PolicyBracket

@typing.final
class RetryPolicy(google.protobuf.message.Message):
    """Retry policy: [ExceptionType -> retry: 3, backoff: 60s]"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    EXCEPTION_TYPES_FIELD_NUMBER: builtins.int
    MAX_RETRIES_FIELD_NUMBER: builtins.int
    BACKOFF_FIELD_NUMBER: builtins.int
    max_retries: builtins.int
    @property
    def exception_types(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """Empty = catch all"""

    @property
    def backoff(self) -> Global___Duration: ...
    def __init__(
        self,
        *,
        exception_types: collections.abc.Iterable[builtins.str] | None = ...,
        max_retries: builtins.int = ...,
        backoff: Global___Duration | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["backoff", b"backoff"]) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "backoff",
            b"backoff",
            "exception_types",
            b"exception_types",
            "max_retries",
            b"max_retries",
        ],
    ) -> None: ...

Global___RetryPolicy: typing_extensions.TypeAlias = RetryPolicy

@typing.final
class TimeoutPolicy(google.protobuf.message.Message):
    """Timeout policy: [timeout: 2m]"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TIMEOUT_FIELD_NUMBER: builtins.int
    @property
    def timeout(self) -> Global___Duration: ...
    def __init__(
        self,
        *,
        timeout: Global___Duration | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["timeout", b"timeout"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["timeout", b"timeout"]) -> None: ...

Global___TimeoutPolicy: typing_extensions.TypeAlias = TimeoutPolicy

@typing.final
class Duration(google.protobuf.message.Message):
    """Duration value (e.g., 30s, 2m, 1h)"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SECONDS_FIELD_NUMBER: builtins.int
    seconds: builtins.int
    def __init__(
        self,
        *,
        seconds: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["seconds", b"seconds"]) -> None: ...

Global___Duration: typing_extensions.TypeAlias = Duration

@typing.final
class Span(google.protobuf.message.Message):
    """-----------------------------------------------------------------------------
    Source Location
    -----------------------------------------------------------------------------

    Source span for error reporting
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    START_LINE_FIELD_NUMBER: builtins.int
    START_COL_FIELD_NUMBER: builtins.int
    END_LINE_FIELD_NUMBER: builtins.int
    END_COL_FIELD_NUMBER: builtins.int
    start_line: builtins.int
    start_col: builtins.int
    end_line: builtins.int
    end_col: builtins.int
    def __init__(
        self,
        *,
        start_line: builtins.int = ...,
        start_col: builtins.int = ...,
        end_line: builtins.int = ...,
        end_col: builtins.int = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "end_col",
            b"end_col",
            "end_line",
            b"end_line",
            "start_col",
            b"start_col",
            "start_line",
            b"start_line",
        ],
    ) -> None: ...

Global___Span: typing_extensions.TypeAlias = Span



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: ast.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'ast.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\tast.proto\x12\x0bwaymark.ast\"6\n\x07Program\x12+\n\tfunctions\x18\x01 \x03(\x0b\x32\x18.waymark.ast.FunctionDef\"\x7f\n\x0b\x46unctionDef\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1f\n\x02io\x18\x02 \x01(\x0b\x32\x13.waymark.ast.IoDecl\x12 \n\x04\x62ody\x18\x03 \x01(\x0b\x32\x12.waymark.ast.Block\x12\x1f\n\x04span\x18\x04 \x01(\x0b\x32\x11.waymark.ast.Span\"J\n\x06IoDecl\x12\x0e\n\x06inputs\x18\x01 \x03(\t\x12\x0f\n\x07outputs\x18\x02 \x03(\t\x12\x1f\n\x04span\x18\x03 \x01(\x0b\x32\x11.waymark.ast.Span\"T\n\x05\x42lock\x12*\n\nstatements\x18\x01 \x03(\x0b\x32\x16.waymark.ast.Statement\x12\x1f\n\x04span\x18\x02 \x01(\x0b\x32\x11.waymark.ast.Span\"\xa0\x05\n\tStatement\x12-\n\nassignment\x18\x01 \x01(\x0b\x32\x17.waymark.ast.AssignmentH\x00\x12.\n\x0b\x61\x63tion_call\x18\x02 \x01(\x0b\x32\x17.waymark.ast.ActionCallH\x00\x12\x32\n\rspread_action\x18\x03 \x01(\x0b\x32\x19.waymark.ast.SpreadActionH\x00\x12\x34\n\x0eparallel_block\x18\x04 \x01(\x0b\x32\x1a.waymark.ast.ParallelBlockH\x00\x12(\n\x08\x66or_loop\x18\x05 \x01(\x0b\x32\x14.waymark.ast.ForLoopH\x00\x12/\n\x0b\x63onditional\x18\x06 \x01(\x0b\x32\x18.waymark.ast.ConditionalH\x00\x12,\n\ntry_except\x18\x07 \x01(\x0b\x32\x16.waymark.ast.TryExceptH\x00\x12.\n\x0breturn_stmt\x18\x08 \x01(\x0b\x32\x17.waymark.ast.ReturnStmtH\x00\x12*\n\texpr_stmt\x18\t \x01(\x0b\x32\x15.waymark.ast.ExprStmtH\x00\x12,\n\nbreak_stmt\x18\x0b \x01(\x0b\x32\x16.waymark.ast.BreakStmtH\x00\x12\x32\n\rcontinue_stmt\x18\x0c \x01(\x0b\x32\x19.waymark.ast.ContinueStmtH\x00\x12,\n\nwhile_loop\x18\r \x01(\x0b\x32\x16.waymark.ast.WhileLoopH\x00\x12,\n\nsleep_stmt\x18\x0e \x01(\x0b\x32\x16.waymark.ast.SleepStmtH\x00\x12\x1f\n\x04span\x18\n \x01(\x0b\x32\x11.waymark.ast.SpanB\x06\n\x04kind\"\x0b\n\tBreakStmt\"\x0e\n\x0c\x43ontinueStmt\"?\n\nAssignment\x12\x0f\n\x07targets\x18\x01 \x03(\t\x12 \n\x05value\x18\x02 \x01(\x0b\x32\x11.waymark.ast.Expr\"\x9d\x01\n\nActionCall\x12\x13\n\x0b\x61\x63tion_name\x18\x01 \x01(\t\x12\"\n\x06kwargs\x18\x02 \x03(\x0b\x32\x12.waymark.ast.Kwarg\x12,\n\x08policies\x18\x03 \x03(\x0b\x32\x1a.waymark.ast.PolicyBracket\x12\x18\n\x0bmodule_name\x18\x04 \x01(\tH\x00\x88\x01\x01\x42\x0e\n\x0c_module_name\"p\n\x0cSpreadAction\x12%\n\ncollection\x18\x01 \x01(\x0b\x32\x11.waymark.ast.Expr\x12\x10\n\x08loop_var\x18\x02 \x01(\t\x12\'\n\x06\x61\x63tion\x18\x03 \x01(\x0b\x32\x17.waymark.ast.ActionCall\"1\n\rParallelBlock\x12 \n\x05\x63\x61lls\x18\x01 \x03(\x0b\x32\x11.waymark.ast.Call\"h\n\x04\x43\x61ll\x12)\n\x06\x61\x63tion\x18\x01 \x01(\x0b\x32\x17.waymark.ast.ActionCallH\x00\x12-\n\x08\x66unction\x18\x02 \x01(\x0b\x32\x19.waymark.ast.FunctionCallH\x00\x42\x06\n\x04kind\"i\n\x07\x46orLoop\x12\x11\n\tloop_vars\x18\x01 \x03(\t\x12#\n\x08iterable\x18\x02 \x01(\x0b\x32\x11.waymark.ast.Expr\x12&\n\nblock_body\x18\x03 \x01(\x0b\x32\x12.waymark.ast.Block\"Y\n\tWhileLoop\x12$\n\tcondition\x18\x01 \x01(\x0b\x32\x11.waymark.ast.Expr\x12&\n\nblock_body\x18\x02 \x01(\x0b\x32\x12.waymark.ast.Block\"\xaa\x01\n\x0b\x43onditional\x12(\n\tif_branch\x18\x01 \x01(\x0b\x32\x15.waymark.ast.IfBranch\x12.\n\relif_branches\x18\x02 \x03(\x0b\x32\x17.waymark.ast.ElifBranch\x12\x31\n\x0b\x65lse_branch\x18\x03 \x01(\x0b\x32\x17.waymark.ast.ElseBranchH\x00\x88\x01\x01\x42\x0e\n\x0c_else_branch\"y\n\x08IfBranch\x12$\n\tcondition\x18\x01 \x01(\x0b\x32\x11.waymark.ast.Expr\x12\x1f\n\x04span\x18\x02 \x01(\x0b\x32\x11.waymark.ast.Span\x12&\n\nblock_body\x18\x03 \x01(\x0b\x32\x12.waymark.ast.Block\"{\n\nElifBranch\x12$\n\tcondition\x18\x01 \x01(\x0b\x32\x11.waymark.ast.Expr\x12\x1f\n\x04span\x18\x02 \x01(\x0b\x32\x11.waymark.ast.Span\x12&\n\nblock_body\x18\x03 \x01(\x0b\x32\x12.waymark.ast.Block\"U\n\nElseBranch\x12\x1f\n\x04span\x18\x01 \x01(\x0b\x32\x11.waymark.ast.Span\x12&\n\nblock_body\x18\x02 \x01(\x0b\x32\x12.waymark.ast.Block\"`\n\tTryExcept\x12,\n\x08handlers\x18\x02 \x03(\x0b\x32\x1a.waymark.ast.ExceptHandler\x12%\n\ttry_block\x18\x03 \x01(\x0b\x32\x12.waymark.ast.Block\"\x9f\x01\n\rExceptHandler\x12\x17\n\x0f\x65xception_types\x18\x01 \x03(\t\x12\x1f\n\x04span\x18\x02 \x01(\x0b\x32\x11.waymark.ast.Span\x12&\n\nblock_body\x18\x03 \x01(\x0b\x32\x12.waymark.ast.Block\x12\x1a\n\rexception_var\x18\x04 \x01(\tH\x00\x88\x01\x01\x42\x10\n\x0e_exception_var\"=\n\nReturnStmt\x12%\n\x05value\x18\x01 \x01(\x0b\x32\x11.waymark.ast.ExprH\x00\x88\x01\x01\x42\x08\n\x06_value\"+\n\x08\x45xprStmt\x12\x1f\n\x04\x65xpr\x18\x01 \x01(\x0b\x32\x11.waymark.ast.Expr\"B\n\tSleepStmt\x12(\n\x08\x64uration\x18\x01 \x01(\x0b\x32\x11.waymark.ast.ExprH\x00\x88\x01\x01\x42\x0b\n\t_duration\"\xc1\x04\n\x04\x45xpr\x12\'\n\x07literal\x18\x01 \x01(\x0b\x32\x14.waymark.ast.LiteralH\x00\x12)\n\x08variable\x18\x02 \x01(\x0b\x32\x15.waymark.ast.VariableH\x00\x12*\n\tbinary_op\x18\x03 \x01(\x0b\x32\x15.waymark.ast.BinaryOpH\x00\x12(\n\x08unary_op\x18\x04 \x01(\x0b\x32\x14.waymark.ast.UnaryOpH\x00\x12%\n\x04list\x18\x05 \x01(\x0b\x32\x15.waymark.ast.ListExprH\x00\x12%\n\x04\x64ict\x18\x06 \x01(\x0b\x32\x15.waymark.ast.DictExprH\x00\x12)\n\x05index\x18\x07 \x01(\x0b\x32\x18.waymark.ast.IndexAccessH\x00\x12%\n\x03\x64ot\x18\x08 \x01(\x0b\x32\x16.waymark.ast.DotAccessH\x00\x12\x32\n\rfunction_call\x18\t \x01(\x0b\x32\x19.waymark.ast.FunctionCallH\x00\x12.\n\x0b\x61\x63tion_call\x18\n \x01(\x0b\x32\x17.waymark.ast.ActionCallH\x00\x12\x32\n\rparallel_expr\x18\x0c \x01(\x0b\x32\x19.waymark.ast.ParallelExprH\x00\x12.\n\x0bspread_expr\x18\r \x01(\x0b\x32\x17.waymark.ast.SpreadExprH\x00\x12\x1f\n\x04span\x18\x0b \x01(\x0b\x32\x11.waymark.ast.SpanB\x06\n\x04kind\"\x7f\n\x07Literal\x12\x13\n\tint_value\x18\x01 \x01(\x03H\x00\x12\x15\n\x0b\x66loat_value\x18\x02 \x01(\x01H\x00\x12\x16\n\x0cstring_value\x18\x03 \x01(\tH\x00\x12\x14\n\nbool_value\x18\x04 \x01(\x08H\x00\x12\x11\n\x07is_none\x18\x05 \x01(\x08H\x00\x42\x07\n\x05value\"\x18\n\x08Variable\x12\x0c\n\x04name\x18\x01 \x01(\t\"v\n\x08\x42inaryOp\x12\x1f\n\x04left\x18\x01 \x01(\x0b\x32\x11.waymark.ast.Expr\x12\'\n\x02op\x18\x02 \x01(\x0e\x32\x1b.waymark.ast.BinaryOperator\x12 \n\x05right\x18\x03 \x01(\x0b\x32\x11.waymark.ast.Expr\"U\n\x07UnaryOp\x12&\n\x02op\x18\x01 \x01(\x0e\x32\x1a.waymark.ast.UnaryOperator\x12\"\n\x07operand\x18\x02 \x01(\x0b\x32\x11.waymark.ast.Expr\"/\n\x08ListExpr\x12#\n\x08\x65lements\x18\x01 \x03(\x0b\x32\x11.waymark.ast.Expr\"3\n\x08\x44ictExpr\x12\'\n\x07\x65ntries\x18\x01 \x03(\x0b\x32\x16.waymark.ast.DictEntry\"M\n\tDictEntry\x12\x1e\n\x03key\x18\x01 \x01(\x0b\x32\x11.waymark.ast.Expr\x12 \n\x05value\x18\x02 \x01(\x0b\x32\x11.waymark.ast.Expr\"R\n\x0bIndexAccess\x12!\n\x06object\x18\x01 \x01(\x0b\x32\x11.waymark.ast.Expr\x12 \n\x05index\x18\x02 \x01(\x0b\x32\x11.waymark.ast.Expr\"A\n\tDotAccess\x12!\n\x06object\x18\x01 \x01(\x0b\x32\x11.waymark.ast.Expr\x12\x11\n\tattribute\x18\x02 \x01(\t\"\x97\x01\n\x0c\x46unctionCall\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1f\n\x04\x61rgs\x18\x02 \x03(\x0b\x32\x11.waymark.ast.Expr\x12\"\n\x06kwargs\x18\x03 \x03(\x0b\x32\x12.waymark.ast.Kwarg\x12\x34\n\x0fglobal_function\x18\x04 \x01(\x0e\x32\x1b.waymark.ast.GlobalFunction\"7\n\x05Kwarg\x12\x0c\n\x04name\x18\x01 \x01(\t\x12 \n\x05value\x18\x02 \x01(\x0b\x32\x11.waymark.ast.Expr\"0\n\x0cParallelExpr\x12 \n\x05\x63\x61lls\x18\x01 \x03(\x0b\x32\x11.waymark.ast.Call\"n\n\nSpreadExpr\x12%\n\ncollection\x18\x01 \x01(\x0b\x32\x11.waymark.ast.Expr\x12\x10\n\x08loop_var\x18\x02 \x01(\t\x12\'\n\x06\x61\x63tion\x18\x03 \x01(\x0b\x32\x17.waymark.ast.ActionCall\"q\n\rPolicyBracket\x12)\n\x05retry\x18\x01 \x01(\x0b\x32\x18.waymark.ast.RetryPolicyH\x00\x12-\n\x07timeout\x18\x02 \x01(\x0b\x32\x1a.waymark.ast.TimeoutPolicyH\x00\x42\x06\n\x04kind\"c\n\x0bRetryPolicy\x12\x17\n\x0f\x65xception_types\x18\x01 \x03(\t\x12\x13\n\x0bmax_retries\x18\x02 \x01(\r\x12&\n\x07\x62\x61\x63koff\x18\x03 \x01(\x0b\x32\x15.waymark.ast.Duration\"7\n\rTimeoutPolicy\x12&\n\x07timeout\x18\x01 \x01(\x0b\x32\x15.waymark.ast.Duration\"\x1b\n\x08\x44uration\x12\x0f\n\x07seconds\x18\x01 \x01(\x04\"P\n\x04Span\x12\x12\n\nstart_line\x18\x01 \x01(\r\x12\x11\n\tstart_col\x18\x02 \x01(\r\x12\x10\n\x08\x65nd_line\x18\x03 \x01(\r\x12\x0f\n\x07\x65nd_col\x18\x04 \x01(\r*\xdc\x02\n\x0e\x42inaryOperator\x12\x19\n\x15\x42INARY_OP_UNSPECIFIED\x10\x00\x12\x11\n\rBINARY_OP_ADD\x10\x01\x12\x11\n\rBINARY_OP_SUB\x10\x02\x12\x11\n\rBINARY_OP_MUL\x10\x03\x12\x11\n\rBINARY_OP_DIV\x10\x04\x12\x17\n\x13\x42INARY_OP_FLOOR_DIV\x10\x05\x12\x11\n\rBINARY_OP_MOD\x10\x06\x12\x10\n\x0c\x42INARY_OP_EQ\x10\n\x12\x10\n\x0c\x42INARY_OP_NE\x10\x0b\x12\x10\n\x0c\x42INARY_OP_LT\x10\x0c\x12\x10\n\x0c\x42INARY_OP_LE\x10\r\x12\x10\n\x0c\x42INARY_OP_GT\x10\x0e\x12\x10\n\x0c\x42INARY_OP_GE\x10\x0f\x12\x10\n\x0c\x42INARY_OP_IN\x10\x10\x12\x14\n\x10\x42INARY_OP_NOT_IN\x10\x11\x12\x11\n\rBINARY_OP_AND\x10\x14\x12\x10\n\x0c\x42INARY_OP_OR\x10\x15*M\n\rUnaryOperator\x12\x18\n\x14UNARY_OP_UNSPECIFIED\x10\x00\x12\x10\n\x0cUNARY_OP_NEG\x10\x01\x12\x10\n\x0cUNARY_OP_NOT\x10\x02*\xa5\x01\n\x0eGlobalFunction\x12\x1f\n\x1bGLOBAL_FUNCTION_UNSPECIFIED\x10\x00\x12\x19\n\x15GLOBAL_FUNCTION_RANGE\x10\x01\x12\x17\n\x13GLOBAL_FUNCTION_LEN\x10\x02\x12\x1d\n\x19GLOBAL_FUNCTION_ENUMERATE\x10\x03\x12\x1f\n\x1bGLOBAL_FUNCTION_ISEXCEPTION\x10\x04\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ast_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_BINARYOPERATOR']._serialized_start=4747
  _globals['_BINARYOPERATOR']._serialized_end=5095
  _globals['_UNARYOPERATOR']._serialized_start=5097
  _globals['_UNARYOPERATOR']._serialized_end=5174
  _globals['_GLOBALFUNCTION']._serialized_start=5177
  _globals['_GLOBALFUNCTION']._serialized_end=5342
  _globals['_PROGRAM']._serialized_start=26
  _globals['_PROGRAM']._serialized_end=80
  _globals['_FUNCTIONDEF']._serialized_start=82
  _globals['_FUNCTIONDEF']._serialized_end=209
  _globals['_IODECL']._serialized_start=211
  _globals['_IODECL']._serialized_end=285
  _globals['_BLOCK']._serialized_start=287
  _globals['_BLOCK']._serialized_end=371
  _globals['_STATEMENT']._serialized_start=374
  _globals['_STATEMENT']._serialized_end=1046
  _globals['_BREAKSTMT']._serialized_start=1048
  _globals['_BREAKSTMT']._serialized_end=1059
  _globals['_CONTINUESTMT']._serialized_start=1061
  _globals['_CONTINUESTMT']._serialized_end=1075
  _globals['_ASSIGNMENT']._serialized_start=1077
  _globals['_ASSIGNMENT']._serialized_end=1140
  _globals['_ACTIONCALL']._serialized_start=1143
  _globals['_ACTIONCALL']._serialized_end=1300
  _globals['_SPREADACTION']._serialized_start=1302
  _globals['_SPREADACTION']._serialized_end=1414
  _globals['_PARALLELBLOCK']._serialized_start=1416
  _globals['_PARALLELBLOCK']._serialized_end=1465
  _globals['_CALL']._serialized_start=1467
  _globals['_CALL']._serialized_end=1571
  _globals['_FORLOOP']._serialized_start=1573
  _globals['_FORLOOP']._serialized_end=1678
  _globals['_WHILELOOP']._serialized_start=1680
  _globals['_WHILELOOP']._serialized_end=1769
  _globals['_CONDITIONAL']._serialized_start=1772
  _globals['_CONDITIONAL']._serialized_end=1942
  _globals['_IFBRANCH']._serialized_start=1944
  _globals['_IFBRANCH']._serialized_end=2065
  _globals['_ELIFBRANCH']._serialized_start=2067
  _globals['_ELIFBRANCH']._serialized_end=2190
  _globals['_ELSEBRANCH']._serialized_start=2192
  _globals['_ELSEBRANCH']._serialized_end=2277
  _globals['_TRYEXCEPT']._serialized_start=2279
  _globals['_TRYEXCEPT']._serialized_end=2375
  _globals['_EXCEPTHANDLER']._serialized_start=2378
  _globals['_EXCEPTHANDLER']._serialized_end=2537
  _globals['_RETURNSTMT']._serialized_start=2539
  _globals['_RETURNSTMT']._serialized_end=2600
  _globals['_EXPRSTMT']._serialized_start=2602
  _globals['_EXPRSTMT']._serialized_end=2645
  _globals['_SLEEPSTMT']._serialized_start=2647
  _globals['_SLEEPSTMT']._serialized_end=2713
  _globals['_EXPR']._serialized_start=2716
  _globals['_EXPR']._serialized_end=3293
  _globals['_LITERAL']._serialized_start=3295
  _globals['_LITERAL']._serialized_end=3422
  _globals['_VARIABLE']._serialized_start=3424
  _globals['_VARIABLE']._serialized_end=3448
  _globals['_BINARYOP']._serialized_start=3450
  _globals['_BINARYOP']._serialized_end=3568
  _globals['_UNARYOP']._serialized_start=3570
  _globals['_UNARYOP']._serialized_end=3655
  _globals['_LISTEXPR']._serialized_start=3657
  _globals['_LISTEXPR']._serialized_end=3704
  _globals['_DICTEXPR']._serialized_start=3706
  _globals['_DICTEXPR']._serialized_end=3757
  _globals['_DICTENTRY']._serialized_start=3759
  _globals['_DICTENTRY']._serialized_end=3836
  _globals['_INDEXACCESS']._serialized_start=3838
  _globals['_INDEXACCESS']._serialized_end=3920
  _globals['_DOTACCESS']._serialized_start=3922
  _globals['_DOTACCESS']._serialized_end=3987
  _globals['_FUNCTIONCALL']._serialized_start=3990
  _globals['_FUNCTIONCALL']._serialized_end=4141
  _globals['_KWARG']._serialized_start=4143
  _globals['_KWARG']._serialized_end=4198
  _globals['_PARALLELEXPR']._serialized_start=4200
  _globals['_PARALLELEXPR']._serialized_end=4248
  _globals['_SPREADEXPR']._serialized_start=4250
  _globals['_SPREADEXPR']._serialized_end=4360
  _globals['_POLICYBRACKET']._serialized_start=4362
  _globals['_POLICYBRACKET']._serialized_end=4475
  _globals['_RETRYPOLICY']._serialized_start=4477
  _globals['_RETRYPOLICY']._serialized_end=4576
  _globals['_TIMEOUTPOLICY']._serialized_start=4578
  _globals['_TIMEOUTPOLICY']._serialized_end=4633
  _globals['_DURATION']._serialized_start=4635
  _globals['_DURATION']._serialized_end=4662
  _globals['_SPAN']._serialized_start=4664
  _globals['_SPAN']._serialized_end=4744
# @@protoc_insertion_point(module_scope)



"""
Workflow base class and registration decorator.

This module provides the foundation for defining workflows that can be
compiled to IR and executed by the Waymark runtime.
"""

import hashlib
import inspect
import os
from dataclasses import dataclass
from datetime import timedelta
from functools import wraps
from threading import RLock
from types import UnionType
from typing import (
    Any,
    Awaitable,
    ClassVar,
    Optional,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from proto import ast_pb2 as ir
from proto import messages_pb2 as pb2

from . import bridge
from .actions import deserialize_result_payload
from .ir_builder import build_workflow_ir
from .logger import configure as configure_logger
from .serialization import build_arguments_from_kwargs
from .workflow_runtime import WorkflowNodeResult, _coerce_value

logger = configure_logger("waymark.workflow")

TWorkflow = TypeVar("TWorkflow", bound="Workflow")
TResult = TypeVar("TResult")


@dataclass(frozen=True)
class RetryPolicy:
    """Retry policy for action execution.

    Maps to IR RetryPolicy: [ExceptionType -> retry: N, backoff: Xs]

    Args:
        attempts: Maximum number of retry attempts.
        exception_types: List of exception type names to retry on. Empty = catch all.
        backoff_seconds: Constant backoff duration between retries in seconds.
    """

    attempts: Optional[int] = None
    exception_types: Optional[list[str]] = None
    backoff_seconds: Optional[float] = None


class Workflow:
    """Base class for workflow definitions."""

    name: ClassVar[Optional[str]] = None
    """Human-friendly identifier. Override to pin the registry key; defaults to lowercase class name."""

    version: ClassVar[Optional[str]] = None
    """Optional workflow version identifier. Defaults to IR hash when unset."""

    concurrent: ClassVar[bool] = False
    """When True, downstream engines may respect DAG-parallel execution; False preserves sequential semantics."""

    _workflow_ir: ClassVar[Optional[ir.Program]] = None
    _ir_lock: ClassVar[RLock] = RLock()
    _workflow_version_id: ClassVar[Optional[str]] = None

    async def run(
        self, *args: Any, _blocking: bool = True, _priority: Optional[int] = None, **kwargs: Any
    ) -> Any:
        raise NotImplementedError

    @classmethod
    def _normalize_run_inputs(cls, args: tuple[Any, ...], kwargs: dict[str, Any]) -> dict[str, Any]:
        try:
            run_impl = cls.__workflow_run_impl__  # type: ignore[attr-defined]
        except AttributeError:
            run_impl = cls.run
        sig = inspect.signature(run_impl)
        params = list(sig.parameters.keys())[1:]  # Skip 'self'

        normalized = dict(kwargs)
        for i, arg in enumerate(args):
            if i < len(params):
                normalized[params[i]] = arg

        bound = sig.bind_partial(None, **normalized)
        bound.apply_defaults()
        return {key: value for key, value in bound.arguments.items() if key != "self"}

    @classmethod
    def _build_initial_context(
        cls, args: tuple[Any, ...], kwargs: dict[str, Any]
    ) -> pb2.WorkflowArguments:
        initial_kwargs = cls._normalize_run_inputs(args, kwargs)
        return build_arguments_from_kwargs(initial_kwargs)

    async def run_action(
        self,
        awaitable: Awaitable[TResult],
        *,
        retry: Optional[RetryPolicy] = None,
        timeout: Optional[float | int | timedelta] = None,
    ) -> TResult:
        """Helper that simply awaits the provided action coroutine.

        The retry and timeout arguments are consumed by the workflow compiler
        (IR builder) rather than the runtime execution path.

        Args:
            awaitable: The action coroutine to execute.
            retry: Retry policy including max attempts, exception types, and backoff.
            timeout: Timeout duration in seconds (or timedelta).
        """
        # Parameters are intentionally unused at runtime; the workflow compiler
        # inspects the AST to record them.
        del retry, timeout
        return await awaitable

    @classmethod
    def short_name(cls) -> str:
        if cls.name:
            return cls.name
        return cls.__name__.lower()

    @classmethod
    def workflow_ir(cls) -> ir.Program:
        """Build and cache the IR program for this workflow."""
        if cls._workflow_ir is None:
            with cls._ir_lock:
                if cls._workflow_ir is None:
                    cls._workflow_ir = build_workflow_ir(cls)
        return cls._workflow_ir

    @classmethod
    def _build_registration_payload(
        cls,
        initial_context: Optional[pb2.WorkflowArguments] = None,
        priority: Optional[int] = None,
    ) -> pb2.WorkflowRegistration:
        """Build a registration payload with the serialized IR."""
        program = cls.workflow_ir()

        # Serialize IR to bytes
        ir_bytes = program.SerializeToString()
        ir_hash = hashlib.sha256(ir_bytes).hexdigest()

        workflow_version = cls.version or ir_hash

        message = pb2.WorkflowRegistration(
            workflow_name=cls.short_name(),
            ir=ir_bytes,
            ir_hash=ir_hash,
            workflow_version=workflow_version,
            concurrent=cls.concurrent,
        )

        if initial_context:
            message.initial_context.CopyFrom(initial_context)

        if priority is not None:
            message.priority = priority

        return message


class WorkflowRegistry:
    """Registry of workflow definitions keyed by workflow name."""

    def __init__(self) -> None:
        self._workflows: dict[str, type[Workflow]] = {}
        self._lock = RLock()

    def register(self, name: str, workflow_cls: type[Workflow]) -> None:
        with self._lock:
            if name in self._workflows:
                raise ValueError(f"workflow '{name}' already registered")
            self._workflows[name] = workflow_cls

    def get(self, name: str) -> Optional[type[Workflow]]:
        with self._lock:
            return self._workflows.get(name)

    def names(self) -> list[str]:
        with self._lock:
            return sorted(self._workflows.keys())

    def reset(self) -> None:
        with self._lock:
            self._workflows.clear()


workflow_registry = WorkflowRegistry()


def workflow(cls: type[TWorkflow]) -> type[TWorkflow]:
    """Decorator that registers workflow classes and caches their IR."""

    if not issubclass(cls, Workflow):
        raise TypeError("workflow decorator requires Workflow subclasses")
    run_impl = cls.run
    if not inspect.iscoroutinefunction(run_impl):
        raise TypeError("workflow run() must be defined with 'async def'")

    @wraps(run_impl)
    async def run_public(
        self: Workflow,
        *args: Any,
        _blocking: bool = True,
        _priority: Optional[int] = None,
        **kwargs: Any,
    ) -> Any:
        if _running_under_pytest():
            logger.debug(
                "pytest run: workflow=%s in_memory=%s",
                cls.short_name(),
                os.environ.get("WAYMARK_BRIDGE_IN_MEMORY"),
            )
            _enable_in_memory_broker()
            initial_context = cls._build_initial_context(args, kwargs)
            payload = cls._build_registration_payload(initial_context, priority=_priority)
            result_bytes = await bridge.execute_workflow(payload.SerializeToString())
            return _deserialize_workflow_result(cls, result_bytes)

        initial_context = cls._build_initial_context(args, kwargs)

        payload = cls._build_registration_payload(initial_context, priority=_priority)
        run_result = await bridge.run_instance(payload.SerializeToString())
        cls._workflow_version_id = run_result.workflow_version_id

        if not _blocking:
            return run_result.workflow_instance_id

        if _skip_wait_for_instance():
            logger.info(
                "Skipping wait_for_instance for workflow %s due to WAYMARK_SKIP_WAIT_FOR_INSTANCE",
                cls.short_name(),
            )
            return None
        result_bytes = await bridge.wait_for_instance(
            instance_id=run_result.workflow_instance_id,
            poll_interval_secs=1.0,
        )
        if result_bytes is None:
            raise TimeoutError(
                f"workflow instance {run_result.workflow_instance_id} did not complete"
            )
        return _deserialize_workflow_result(cls, result_bytes)

    cls.__workflow_run_impl__ = run_impl
    cls.run = run_public  # type: ignore[assignment]
    workflow_registry.register(cls.short_name(), cls)
    return cls


def _running_under_pytest() -> bool:
    return bool(os.environ.get("PYTEST_CURRENT_TEST"))


def _skip_wait_for_instance() -> bool:
    value = os.environ.get("WAYMARK_SKIP_WAIT_FOR_INSTANCE")
    if not value:
        return False
    return value.strip().lower() not in {"0", "false", "no"}


def _enable_in_memory_broker() -> None:
    os.environ.setdefault("WAYMARK_BRIDGE_IN_MEMORY", "1")


def _deserialize_workflow_result(
    workflow_cls: type[Workflow],
    result_bytes: bytes,
) -> Any:
    arguments = pb2.WorkflowArguments()
    arguments.ParseFromString(result_bytes)
    result = deserialize_result_payload(arguments)
    if result.error:
        raise RuntimeError(f"workflow failed: {result.error}")

    value = result.result

    # Unwrap WorkflowNodeResult if present (internal worker representation)
    if isinstance(value, WorkflowNodeResult):
        variables = value.variables
        program = workflow_cls.workflow_ir()
        value = None
        if program.functions:
            outputs = list(program.functions[0].io.outputs)
            for name in outputs:
                if name in variables:
                    value = variables[name]
                    break
        if value is None:
            if "result" in variables:
                value = variables["result"]
            elif len(variables) == 1:
                value = next(iter(variables.values()))
            else:
                value = variables
    target_type = _resolve_return_type(workflow_cls)
    if target_type is None:
        return value
    return _coerce_result_value(value, target_type)


def _resolve_return_type(workflow_cls: type[Workflow]) -> Optional[type]:
    try:
        run_impl = workflow_cls.__workflow_run_impl__  # type: ignore[attr-defined]
    except AttributeError:
        run_impl = workflow_cls.run
    try:
        type_hints = get_type_hints(run_impl)
    except Exception:
        return None
    target_type = type_hints.get("return")
    if target_type is None or target_type is Any:
        return None
    return target_type


def _coerce_result_value(value: Any, target_type: type) -> Any:
    origin = get_origin(target_type)
    if origin is UnionType or origin is Union:
        for arg in get_args(target_type):
            if arg is type(None) and value is None:
                return None
            try:
                coerced = _coerce_value(value, arg)
            except Exception:
                continue
            if coerced is not value:
                return coerced
            if isinstance(arg, type) and isinstance(value, arg):
                return value
        return value
    try:
        return _coerce_value(value, target_type)
    except Exception:
        return value



import inspect
from dataclasses import dataclass
from functools import wraps
from typing import Any, Callable, Optional, TypeVar, overload

from proto import messages_pb2 as pb2

from .dependencies import provide_dependencies
from .registry import AsyncAction, registry
from .serialization import dumps, loads

TAsync = TypeVar("TAsync", bound=AsyncAction)


@dataclass
class ActionResultPayload:
    result: Any | None
    error: dict[str, Any] | None


def serialize_result_payload(value: Any) -> pb2.WorkflowArguments:
    """Serialize a successful action result."""
    arguments = pb2.WorkflowArguments()
    entry = arguments.arguments.add()
    entry.key = "result"
    entry.value.CopyFrom(dumps(value))
    return arguments


def serialize_error_payload(_action: str, exc: BaseException) -> pb2.WorkflowArguments:
    """Serialize an error raised during action execution."""
    arguments = pb2.WorkflowArguments()
    entry = arguments.arguments.add()
    entry.key = "error"
    entry.value.CopyFrom(dumps(exc))
    return arguments


def deserialize_result_payload(payload: pb2.WorkflowArguments | None) -> ActionResultPayload:
    """Deserialize WorkflowArguments produced by serialize_result_payload/error."""
    if payload is None:
        return ActionResultPayload(result=None, error=None)
    values = {entry.key: entry.value for entry in payload.arguments}
    if "error" in values:
        error_value = values["error"]
        data = loads(error_value)
        if not isinstance(data, dict):
            raise ValueError("error payload must deserialize to a mapping")
        return ActionResultPayload(result=None, error=data)
    result_value = values.get("result")
    if result_value is None:
        raise ValueError("result payload missing 'result' field")
    return ActionResultPayload(result=loads(result_value), error=None)


@overload
def action(func: TAsync, /) -> TAsync: ...


@overload
def action(*, name: Optional[str] = None) -> Callable[[TAsync], TAsync]: ...


def action(
    func: Optional[TAsync] = None,
    *,
    name: Optional[str] = None,
) -> Callable[[TAsync], TAsync] | TAsync:
    """Decorator for registering async actions.

    Actions decorated with @action will automatically resolve Depend() markers
    when called directly (e.g., during pytest runs where workflows bypass the
    gRPC bridge).
    """

    def decorator(target: TAsync) -> TAsync:
        if not inspect.iscoroutinefunction(target):
            raise TypeError(f"action '{target.__name__}' must be defined with 'async def'")
        action_name = name or target.__name__
        action_module = target.__module__

        @wraps(target)
        async def wrapper(*args: Any, **kwargs: Any) -> Any:
            # Convert positional args to kwargs based on the signature
            sig = inspect.signature(target)
            params = list(sig.parameters.keys())
            for i, arg in enumerate(args):
                if i < len(params):
                    kwargs[params[i]] = arg

            # Resolve dependencies using the same mechanism as execute_action
            async with provide_dependencies(target, kwargs) as call_kwargs:
                return await target(**call_kwargs)

        # Copy over the original function's attributes for introspection
        wrapper.__wrapped__ = target  # type: ignore[attr-defined]
        wrapper.__waymark_action_name__ = action_name  # type: ignore[attr-defined]
        wrapper.__waymark_action_module__ = action_module  # type: ignore[attr-defined]

        # Register the original function (not the wrapper) so execute_action
        # doesn't double-resolve dependencies
        registry.register(action_module, action_name, target)

        return wrapper  # type: ignore[return-value]

    if func is not None:
        return decorator(func)
    return decorator



from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from threading import RLock
from typing import Any, Optional

AsyncAction = Callable[..., Awaitable[Any]]


@dataclass
class _ActionEntry:
    module: str
    name: str
    func: AsyncAction


def _make_key(module: str, name: str) -> str:
    """Create a registry key from module and action name."""
    return f"{module}:{name}"


class ActionRegistry:
    """In-memory registry of user-defined actions.

    Actions are keyed by (module, name), allowing the same action name
    to be used in different modules.
    """

    def __init__(self) -> None:
        self._actions: dict[str, _ActionEntry] = {}
        self._lock = RLock()

    def _source_fingerprint(self, func: AsyncAction) -> tuple[str | None, str | None]:
        func_any: Any = func
        try:
            code = func_any.__code__
        except AttributeError:
            return (None, None)
        try:
            qualname = func_any.__qualname__
        except AttributeError:
            qualname = None
        filename = code.co_filename
        if not isinstance(filename, str):
            filename = None
        if qualname is not None and not isinstance(qualname, str):
            qualname = None
        return (filename, qualname)

    def _is_same_action_definition(self, existing: AsyncAction, new: AsyncAction) -> bool:
        if existing is new:
            return True
        existing_fingerprint = self._source_fingerprint(existing)
        new_fingerprint = self._source_fingerprint(new)
        if existing_fingerprint == (None, None) or new_fingerprint == (None, None):
            return False
        return existing_fingerprint == new_fingerprint

    def register(self, module: str, name: str, func: AsyncAction) -> None:
        """Register an action with its module and name.

        Args:
            module: The Python module containing the action.
            name: The action name (from @action decorator).
            func: The async function to execute.

        Raises:
            ValueError: If an action with the same module:name is already registered
                with a different implementation.
        """
        key = _make_key(module, name)
        with self._lock:
            existing = self._actions.get(key)
            if existing is not None:
                if self._is_same_action_definition(existing.func, func):
                    self._actions[key] = _ActionEntry(module=module, name=name, func=func)
                    return
                raise ValueError(f"action '{module}:{name}' already registered")
            self._actions[key] = _ActionEntry(module=module, name=name, func=func)

    def get(self, module: str, name: str) -> Optional[AsyncAction]:
        """Look up an action by module and name.

        Args:
            module: The Python module containing the action.
            name: The action name.

        Returns:
            The action function if found, None otherwise.
        """
        key = _make_key(module, name)
        with self._lock:
            entry = self._actions.get(key)
            return entry.func if entry else None

    def names(self) -> list[str]:
        """Return all registered action keys (module:name format)."""
        with self._lock:
            return sorted(self._actions.keys())

    def entries(self) -> list[_ActionEntry]:
        """Return all registered action entries."""
        with self._lock:
            return list(self._actions.values())

    def reset(self) -> None:
        """Clear all registered actions."""
        with self._lock:
            self._actions.clear()


registry = ActionRegistry()



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""

import grpc

from proto import messages_pb2 as messages__pb2

GRPC_GENERATED_VERSION = '1.71.2'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + ' but the generated code in messages_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class WorkerBridgeStub(object):
    """=============================================================================
    gRPC Service Definitions
    =============================================================================

    Bidirectional streaming service for worker communication.
    Workers connect and maintain a persistent stream for action dispatch/results.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Attach = channel.stream_stream(
                '/waymark.messages.WorkerBridge/Attach',
                request_serializer=messages__pb2.Envelope.SerializeToString,
                response_deserializer=messages__pb2.Envelope.FromString,
                _registered_method=True)


class WorkerBridgeServicer(object):
    """=============================================================================
    gRPC Service Definitions
    =============================================================================

    Bidirectional streaming service for worker communication.
    Workers connect and maintain a persistent stream for action dispatch/results.
    """

    def Attach(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_WorkerBridgeServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Attach': grpc.stream_stream_rpc_method_handler(
                    servicer.Attach,
                    request_deserializer=messages__pb2.Envelope.FromString,
                    response_serializer=messages__pb2.Envelope.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'waymark.messages.WorkerBridge', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('waymark.messages.WorkerBridge', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class WorkerBridge(object):
    """=============================================================================
    gRPC Service Definitions
    =============================================================================

    Bidirectional streaming service for worker communication.
    Workers connect and maintain a persistent stream for action dispatch/results.
    """

    @staticmethod
    def Attach(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            '/waymark.messages.WorkerBridge/Attach',
            messages__pb2.Envelope.SerializeToString,
            messages__pb2.Envelope.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class WorkflowServiceStub(object):
    """Workflow management service for client operations.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.RegisterWorkflow = channel.unary_unary(
                '/waymark.messages.WorkflowService/RegisterWorkflow',
                request_serializer=messages__pb2.RegisterWorkflowRequest.SerializeToString,
                response_deserializer=messages__pb2.RegisterWorkflowResponse.FromString,
                _registered_method=True)
        self.RegisterWorkflowBatch = channel.unary_unary(
                '/waymark.messages.WorkflowService/RegisterWorkflowBatch',
                request_serializer=messages__pb2.RegisterWorkflowBatchRequest.SerializeToString,
                response_deserializer=messages__pb2.RegisterWorkflowBatchResponse.FromString,
                _registered_method=True)
        self.WaitForInstance = channel.unary_unary(
                '/waymark.messages.WorkflowService/WaitForInstance',
                request_serializer=messages__pb2.WaitForInstanceRequest.SerializeToString,
                response_deserializer=messages__pb2.WaitForInstanceResponse.FromString,
                _registered_method=True)
        self.ExecuteWorkflow = channel.stream_stream(
                '/waymark.messages.WorkflowService/ExecuteWorkflow',
                request_serializer=messages__pb2.WorkflowStreamRequest.SerializeToString,
                response_deserializer=messages__pb2.WorkflowStreamResponse.FromString,
                _registered_method=True)
        self.RegisterSchedule = channel.unary_unary(
                '/waymark.messages.WorkflowService/RegisterSchedule',
                request_serializer=messages__pb2.RegisterScheduleRequest.SerializeToString,
                response_deserializer=messages__pb2.RegisterScheduleResponse.FromString,
                _registered_method=True)
        self.UpdateScheduleStatus = channel.unary_unary(
                '/waymark.messages.WorkflowService/UpdateScheduleStatus',
                request_serializer=messages__pb2.UpdateScheduleStatusRequest.SerializeToString,
                response_deserializer=messages__pb2.UpdateScheduleStatusResponse.FromString,
                _registered_method=True)
        self.DeleteSchedule = channel.unary_unary(
                '/waymark.messages.WorkflowService/DeleteSchedule',
                request_serializer=messages__pb2.DeleteScheduleRequest.SerializeToString,
                response_deserializer=messages__pb2.DeleteScheduleResponse.FromString,
                _registered_method=True)
        self.ListSchedules = channel.unary_unary(
                '/waymark.messages.WorkflowService/ListSchedules',
                request_serializer=messages__pb2.ListSchedulesRequest.SerializeToString,
                response_deserializer=messages__pb2.ListSchedulesResponse.FromString,
                _registered_method=True)


class WorkflowServiceServicer(object):
    """Workflow management service for client operations.
    """

    def RegisterWorkflow(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RegisterWorkflowBatch(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def WaitForInstance(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ExecuteWorkflow(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RegisterSchedule(self, request, context):
        """Schedule management
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateScheduleStatus(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteSchedule(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListSchedules(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_WorkflowServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'RegisterWorkflow': grpc.unary_unary_rpc_method_handler(
                    servicer.RegisterWorkflow,
                    request_deserializer=messages__pb2.RegisterWorkflowRequest.FromString,
                    response_serializer=messages__pb2.RegisterWorkflowResponse.SerializeToString,
            ),
            'RegisterWorkflowBatch': grpc.unary_unary_rpc_method_handler(
                    servicer.RegisterWorkflowBatch,
                    request_deserializer=messages__pb2.RegisterWorkflowBatchRequest.FromString,
                    response_serializer=messages__pb2.RegisterWorkflowBatchResponse.SerializeToString,
            ),
            'WaitForInstance': grpc.unary_unary_rpc_method_handler(
                    servicer.WaitForInstance,
                    request_deserializer=messages__pb2.WaitForInstanceRequest.FromString,
                    response_serializer=messages__pb2.WaitForInstanceResponse.SerializeToString,
            ),
            'ExecuteWorkflow': grpc.stream_stream_rpc_method_handler(
                    servicer.ExecuteWorkflow,
                    request_deserializer=messages__pb2.WorkflowStreamRequest.FromString,
                    response_serializer=messages__pb2.WorkflowStreamResponse.SerializeToString,
            ),
            'RegisterSchedule': grpc.unary_unary_rpc_method_handler(
                    servicer.RegisterSchedule,
                    request_deserializer=messages__pb2.RegisterScheduleRequest.FromString,
                    response_serializer=messages__pb2.RegisterScheduleResponse.SerializeToString,
            ),
            'UpdateScheduleStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateScheduleStatus,
                    request_deserializer=messages__pb2.UpdateScheduleStatusRequest.FromString,
                    response_serializer=messages__pb2.UpdateScheduleStatusResponse.SerializeToString,
            ),
            'DeleteSchedule': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteSchedule,
                    request_deserializer=messages__pb2.DeleteScheduleRequest.FromString,
                    response_serializer=messages__pb2.DeleteScheduleResponse.SerializeToString,
            ),
            'ListSchedules': grpc.unary_unary_rpc_method_handler(
                    servicer.ListSchedules,
                    request_deserializer=messages__pb2.ListSchedulesRequest.FromString,
                    response_serializer=messages__pb2.ListSchedulesResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'waymark.messages.WorkflowService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('waymark.messages.WorkflowService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class WorkflowService(object):
    """Workflow management service for client operations.
    """

    @staticmethod
    def RegisterWorkflow(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/waymark.messages.WorkflowService/RegisterWorkflow',
            messages__pb2.RegisterWorkflowRequest.SerializeToString,
            messages__pb2.RegisterWorkflowResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RegisterWorkflowBatch(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/waymark.messages.WorkflowService/RegisterWorkflowBatch',
            messages__pb2.RegisterWorkflowBatchRequest.SerializeToString,
            messages__pb2.RegisterWorkflowBatchResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def WaitForInstance(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/waymark.messages.WorkflowService/WaitForInstance',
            messages__pb2.WaitForInstanceRequest.SerializeToString,
            messages__pb2.WaitForInstanceResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ExecuteWorkflow(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            '/waymark.messages.WorkflowService/ExecuteWorkflow',
            messages__pb2.WorkflowStreamRequest.SerializeToString,
            messages__pb2.WorkflowStreamResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RegisterSchedule(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/waymark.messages.WorkflowService/RegisterSchedule',
            messages__pb2.RegisterScheduleRequest.SerializeToString,
            messages__pb2.RegisterScheduleResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def UpdateScheduleStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/waymark.messages.WorkflowService/UpdateScheduleStatus',
            messages__pb2.UpdateScheduleStatusRequest.SerializeToString,
            messages__pb2.UpdateScheduleStatusResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DeleteSchedule(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/waymark.messages.WorkflowService/DeleteSchedule',
            messages__pb2.DeleteScheduleRequest.SerializeToString,
            messages__pb2.DeleteScheduleResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListSchedules(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/waymark.messages.WorkflowService/ListSchedules',
            messages__pb2.ListSchedulesRequest.SerializeToString,
            messages__pb2.ListSchedulesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



"""gRPC worker client that executes waymark actions."""

import argparse
import asyncio
import importlib
import logging
import os
import sys
import time
from typing import Any, AsyncIterator, cast

import grpc

from proto import messages_pb2 as pb2
from proto import messages_pb2_grpc as pb2_grpc
from waymark.actions import serialize_error_payload, serialize_result_payload

from . import workflow_runtime
from .logger import configure as configure_logger

LOGGER = configure_logger("waymark.worker")
aio = cast(Any, grpc).aio

DEFAULT_CLEANUP_TIMEOUT_BUFFER_SECONDS = 5


def _cleanup_timeout_buffer_seconds() -> int:
    raw = os.getenv("WAYMARK_PYTHON_CLEANUP_TIMEOUT_BUFFER_SECONDS")
    if raw is None:
        return DEFAULT_CLEANUP_TIMEOUT_BUFFER_SECONDS
    try:
        return max(1, int(raw))
    except ValueError:
        return DEFAULT_CLEANUP_TIMEOUT_BUFFER_SECONDS


def _parse_args(argv: list[str] | None) -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Waymark workflow worker")
    parser.add_argument("--bridge", required=True, help="gRPC address of the Rust bridge")
    parser.add_argument("--worker-id", required=True, type=int, help="Logical worker identifier")
    parser.add_argument(
        "--user-module",
        action="append",
        default=[],
        help="Optional user module(s) to import eagerly",
    )
    return parser.parse_args(argv)


async def _outgoing_stream(
    queue: "asyncio.Queue[pb2.Envelope]", worker_id: int
) -> AsyncIterator[pb2.Envelope]:
    hello = pb2.WorkerHello(worker_id=worker_id)
    envelope = pb2.Envelope(
        delivery_id=0,
        partition_id=0,
        kind=pb2.MessageKind.MESSAGE_KIND_WORKER_HELLO,
        payload=hello.SerializeToString(),
    )
    yield envelope
    try:
        while True:
            message = await queue.get()
            yield message
    except asyncio.CancelledError:  # pragma: no cover - best effort shutdown
        return


async def _send_ack(outgoing: "asyncio.Queue[pb2.Envelope]", envelope: pb2.Envelope) -> None:
    ack = pb2.Ack(acked_delivery_id=envelope.delivery_id)
    ack_envelope = pb2.Envelope(
        delivery_id=envelope.delivery_id,
        partition_id=envelope.partition_id,
        kind=pb2.MessageKind.MESSAGE_KIND_ACK,
        payload=ack.SerializeToString(),
    )
    await outgoing.put(ack_envelope)


async def _handle_dispatch(
    envelope: pb2.Envelope,
    outgoing: "asyncio.Queue[pb2.Envelope]",
) -> None:
    await _send_ack(outgoing, envelope)
    dispatch = pb2.ActionDispatch()
    dispatch.ParseFromString(envelope.payload)
    timeout_seconds = dispatch.timeout_seconds if dispatch.HasField("timeout_seconds") else 0

    # Python-side timeout is a safety net for cleanup only.
    # Rust handles the primary timeout enforcement.
    # Add buffer so Rust always times out first.
    cleanup_timeout_buffer = _cleanup_timeout_buffer_seconds()
    python_timeout = timeout_seconds + cleanup_timeout_buffer if timeout_seconds > 0 else 0

    worker_start = time.perf_counter_ns()
    success = True
    action_name = dispatch.action_name
    execution: workflow_runtime.ActionExecutionResult | None = None
    try:
        if python_timeout > 0:
            execution = await asyncio.wait_for(
                workflow_runtime.execute_action(dispatch), timeout=python_timeout
            )
        else:
            execution = await workflow_runtime.execute_action(dispatch)

        if execution.exception:
            success = False
            response_payload = serialize_error_payload(action_name, execution.exception)
        else:
            response_payload = serialize_result_payload(execution.result)
    except asyncio.TimeoutError:
        # Python-side timeout is just for cleanup - Rust already handled the timeout.
        # Log internally but don't treat as special error type.
        LOGGER.warning(
            "Action %s hit Python cleanup timeout after %ss (Rust already timed out at %ss) "
            "for action_id=%s sequence=%s attempt=%s",
            action_name,
            python_timeout,
            timeout_seconds,
            dispatch.action_id,
            dispatch.sequence,
            dispatch.attempt_number,
        )
        success = False
        # Return generic error - Rust will likely ignore this late response
        error = Exception(
            f"action {action_name} cleanup timeout (Rust-side timeout already triggered)"
        )
        response_payload = serialize_error_payload(action_name, error)
    except Exception as exc:  # noqa: BLE001 - propagate structured errors
        success = False
        response_payload = serialize_error_payload(action_name, exc)
        LOGGER.exception(
            "Action %s failed for action_id=%s sequence=%s",
            action_name,
            dispatch.action_id,
            dispatch.sequence,
        )
    worker_end = time.perf_counter_ns()
    response = pb2.ActionResult(
        action_id=dispatch.action_id,
        success=success,
        worker_start_ns=worker_start,
        worker_end_ns=worker_end,
    )
    response.payload.CopyFrom(response_payload)
    if dispatch.dispatch_token:
        response.dispatch_token = dispatch.dispatch_token
    response_envelope = pb2.Envelope(
        delivery_id=envelope.delivery_id,
        partition_id=envelope.partition_id,
        kind=pb2.MessageKind.MESSAGE_KIND_ACTION_RESULT,
        payload=response.SerializeToString(),
    )
    await outgoing.put(response_envelope)
    LOGGER.debug(
        "Handled action=%s seq=%s attempt=%s success=%s",
        action_name,
        dispatch.sequence,
        dispatch.attempt_number,
        success,
    )


async def _handle_incoming_stream(
    stub: pb2_grpc.WorkerBridgeStub,
    worker_id: int,
    outgoing: "asyncio.Queue[pb2.Envelope]",
) -> None:
    """Process incoming messages, running action dispatches concurrently."""
    pending_tasks: set[asyncio.Task[None]] = set()

    async for envelope in stub.Attach(_outgoing_stream(outgoing, worker_id)):
        kind = envelope.kind
        if kind == pb2.MessageKind.MESSAGE_KIND_ACTION_DISPATCH:
            # Spawn task to handle dispatch concurrently
            task = asyncio.create_task(_handle_dispatch(envelope, outgoing))
            pending_tasks.add(task)
            task.add_done_callback(pending_tasks.discard)
        elif kind == pb2.MessageKind.MESSAGE_KIND_HEARTBEAT:
            LOGGER.debug("Received heartbeat delivery=%s", envelope.delivery_id)
            await _send_ack(outgoing, envelope)
        else:
            LOGGER.warning("Unhandled message kind: %s", kind)
            await _send_ack(outgoing, envelope)

    # Wait for any remaining tasks on stream close
    if pending_tasks:
        await asyncio.gather(*pending_tasks, return_exceptions=True)


async def _run_worker(args: argparse.Namespace) -> None:
    outgoing: "asyncio.Queue[pb2.Envelope]" = asyncio.Queue()
    for module_name in args.user_module:
        if not module_name:
            continue
        LOGGER.info("Preloading user module %s", module_name)
        importlib.import_module(module_name)

    async with aio.insecure_channel(args.bridge) as channel:
        stub = pb2_grpc.WorkerBridgeStub(channel)
        LOGGER.info("Worker %s connected to %s", args.worker_id, args.bridge)
        try:
            await _handle_incoming_stream(stub, args.worker_id, outgoing)
        except aio.AioRpcError as exc:  # pragma: no cover
            status = exc.code()
            LOGGER.error("Worker stream closed: %s", status)
            raise


def main(argv: list[str] | None = None) -> None:
    args = _parse_args(argv)
    logging.basicConfig(level=logging.INFO, format="[worker] %(message)s", stream=sys.stderr)
    try:
        asyncio.run(_run_worker(args))
    except KeyboardInterrupt:  # pragma: no cover - exit quietly on Ctrl+C
        return
    except grpc.RpcError:
        sys.exit(1)


if __name__ == "__main__":
    main()



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import abc
import collections.abc
import typing

import grpc
import grpc.aio

_T = typing.TypeVar("_T")

class _MaybeAsyncIterator(
    collections.abc.AsyncIterator[_T], collections.abc.Iterator[_T], metaclass=abc.ABCMeta
): ...
class _ServicerContext(grpc.ServicerContext, grpc.aio.ServicerContext):  # type: ignore[misc, type-arg]
    ...

GRPC_GENERATED_VERSION: str
GRPC_VERSION: str



"""Runtime helpers for executing actions inside the worker.

This module provides the execution layer for Python workers that receive
action dispatch commands from the Rust scheduler.
"""

import asyncio
import dataclasses
from base64 import b64decode
from dataclasses import dataclass
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from pathlib import Path, PurePath
from typing import Any, Dict, get_args, get_origin, get_type_hints
from uuid import UUID

from pydantic import BaseModel

from proto import messages_pb2 as pb2

from .dependencies import provide_dependencies
from .registry import registry
from .serialization import arguments_to_kwargs


class WorkflowNodeResult(BaseModel):
    """Result from a workflow node execution containing variable bindings."""

    variables: Dict[str, Any]


@dataclass
class ActionExecutionResult:
    """Result of an action execution."""

    result: Any
    exception: BaseException | None = None


def _is_pydantic_model(cls: type) -> bool:
    """Check if a class is a Pydantic BaseModel subclass."""
    try:
        return isinstance(cls, type) and issubclass(cls, BaseModel)
    except TypeError:
        return False


def _is_dataclass_type(cls: type) -> bool:
    """Check if a class is a dataclass."""
    return dataclasses.is_dataclass(cls) and isinstance(cls, type)


def _coerce_primitive(value: Any, target_type: type) -> Any:
    """Coerce a value to a primitive type based on target_type.

    Handles conversion of serialized values (strings, floats) back to their
    native Python types (UUID, datetime, etc.).
    """
    # Handle None
    if value is None:
        return None

    # UUID from string
    if target_type is UUID:
        if isinstance(value, UUID):
            return value
        if isinstance(value, str):
            return UUID(value)
        return value

    # datetime from ISO string
    if target_type is datetime:
        if isinstance(value, datetime):
            return value
        if isinstance(value, str):
            return datetime.fromisoformat(value)
        return value

    # date from ISO string
    if target_type is date:
        if isinstance(value, date):
            return value
        if isinstance(value, str):
            return date.fromisoformat(value)
        return value

    # time from ISO string
    if target_type is time:
        if isinstance(value, time):
            return value
        if isinstance(value, str):
            return time.fromisoformat(value)
        return value

    # timedelta from total seconds
    if target_type is timedelta:
        if isinstance(value, timedelta):
            return value
        if isinstance(value, (int, float)):
            return timedelta(seconds=value)
        return value

    # Decimal from string
    if target_type is Decimal:
        if isinstance(value, Decimal):
            return value
        if isinstance(value, (str, int, float)):
            return Decimal(str(value))
        return value

    # bytes from base64 string
    if target_type is bytes:
        if isinstance(value, bytes):
            return value
        if isinstance(value, str):
            return b64decode(value)
        return value

    # Path from string
    if target_type is Path or target_type is PurePath:
        if isinstance(value, PurePath):
            return value
        if isinstance(value, str):
            return Path(value)
        return value

    return value


# Types that can be coerced from serialized form
COERCIBLE_TYPES = (UUID, datetime, date, time, timedelta, Decimal, bytes, Path, PurePath)


def _coerce_dict_to_model(value: Any, target_type: type) -> Any:
    """Convert a dict to a Pydantic model or dataclass if needed.

    If value is a dict and target_type is a Pydantic model or dataclass,
    instantiate the model with the dict values. Otherwise, return value unchanged.
    """
    if not isinstance(value, dict):
        return value

    if _is_pydantic_model(target_type):
        # Use model_validate for Pydantic v2, fall back to direct instantiation
        model_validate = getattr(target_type, "model_validate", None)
        if model_validate is not None:
            return model_validate(value)
        return target_type(**value)

    if _is_dataclass_type(target_type):
        return target_type(**value)

    return value


def _coerce_value(value: Any, target_type: type) -> Any:
    """Coerce a value to the target type.

    Handles:
    - Primitive types (UUID, datetime, etc.)
    - Pydantic models and dataclasses (from dicts)
    - Generic collections like list[UUID], set[datetime]
    """
    # Handle None
    if value is None:
        return None

    # Check for coercible primitive types
    if isinstance(target_type, type) and issubclass(target_type, COERCIBLE_TYPES):
        return _coerce_primitive(value, target_type)

    # Check for Pydantic models or dataclasses
    if isinstance(value, dict):
        coerced = _coerce_dict_to_model(value, target_type)
        if coerced is not value:
            return coerced

    # Handle generic types like list[UUID], set[datetime]
    origin = get_origin(target_type)
    if origin is not None:
        args = get_args(target_type)

        # Handle list[T]
        if origin is list and isinstance(value, list) and args:
            item_type = args[0]
            return [_coerce_value(item, item_type) for item in value]

        # Handle set[T] (serialized as list)
        if origin is set and isinstance(value, list) and args:
            item_type = args[0]
            return {_coerce_value(item, item_type) for item in value}

        # Handle frozenset[T] (serialized as list)
        if origin is frozenset and isinstance(value, list) and args:
            item_type = args[0]
            return frozenset(_coerce_value(item, item_type) for item in value)

        # Handle tuple[T, ...] (serialized as list)
        if origin is tuple and isinstance(value, (list, tuple)) and args:
            # Variable length tuple like tuple[int, ...]
            if len(args) == 2 and args[1] is ...:
                item_type = args[0]
                return tuple(_coerce_value(item, item_type) for item in value)
            # Fixed length tuple like tuple[int, str, UUID]
            return tuple(
                _coerce_value(item, item_type) for item, item_type in zip(value, args, strict=False)
            )

        # Handle dict[K, V]
        if origin is dict and isinstance(value, dict) and len(args) == 2:
            key_type, val_type = args
            return {
                _coerce_value(k, key_type): _coerce_value(v, val_type) for k, v in value.items()
            }

    return value


def _coerce_kwargs_to_type_hints(handler: Any, kwargs: Dict[str, Any]) -> Dict[str, Any]:
    """Coerce kwargs to expected types based on handler's type hints.

    Handles:
    - Pydantic models and dataclasses (from dicts)
    - Primitive types like UUID, datetime, Decimal, etc.
    - Generic collections like list[UUID], dict[str, datetime]
    """
    try:
        type_hints = get_type_hints(handler)
    except Exception:
        # If we can't get type hints (e.g., forward references), return as-is
        return kwargs

    coerced = {}
    for key, value in kwargs.items():
        if key in type_hints:
            target_type = type_hints[key]
            coerced[key] = _coerce_value(value, target_type)
        else:
            coerced[key] = value

    return coerced


async def execute_action(dispatch: pb2.ActionDispatch) -> ActionExecutionResult:
    """Execute an action based on the dispatch command.

    Args:
        dispatch: The action dispatch command from the Rust scheduler.

    Returns:
        The result of executing the action.
    """
    action_name = dispatch.action_name
    module_name = dispatch.module_name

    module = None
    if module_name:
        import importlib

        module = importlib.import_module(module_name)

    # Get the action handler using both module and name
    handler = registry.get(module_name, action_name)
    if handler is None and module is not None:
        import importlib

        module = importlib.reload(module)
        handler = registry.get(module_name, action_name)
    if handler is None:
        return ActionExecutionResult(
            result=None,
            exception=KeyError(f"action '{module_name}:{action_name}' not registered"),
        )

    # Deserialize kwargs
    kwargs = arguments_to_kwargs(dispatch.kwargs)

    # Coerce dict arguments to Pydantic models or dataclasses based on type hints
    # This is needed because the IR converts model constructor calls to dicts
    kwargs = _coerce_kwargs_to_type_hints(handler, kwargs)

    try:
        async with provide_dependencies(handler, kwargs) as call_kwargs:
            value = handler(**call_kwargs)
            if asyncio.iscoroutine(value):
                value = await value
        return ActionExecutionResult(result=value)
    except Exception as e:
        return ActionExecutionResult(
            result=None,
            exception=e,
        )



"""
Scheduled workflow execution.

This module provides functions for registering workflows to run on a cron
schedule or at fixed intervals.
"""

from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Any, Dict, List, Literal, Optional, Type, Union

from grpc import StatusCode, aio  # type: ignore[attr-defined]

from proto import messages_pb2 as pb2

from .bridge import _workflow_stub, ensure_singleton
from .exceptions import ScheduleAlreadyExistsError
from .workflow import Workflow

ScheduleType = Literal["cron", "interval"]
ScheduleStatus = Literal["active", "paused"]


@dataclass
class ScheduleInfo:
    """Information about a registered schedule."""

    id: str
    workflow_name: str
    schedule_name: str
    schedule_type: ScheduleType
    cron_expression: Optional[str]
    interval_seconds: Optional[int]
    jitter_seconds: Optional[int]
    status: ScheduleStatus
    next_run_at: Optional[datetime]
    last_run_at: Optional[datetime]
    last_instance_id: Optional[str]
    created_at: datetime
    updated_at: datetime
    allow_duplicate: bool


async def schedule_workflow(
    workflow_cls: Type[Workflow],
    *,
    schedule_name: str,
    schedule: Union[str, timedelta],
    jitter: Optional[timedelta] = None,
    inputs: Optional[Dict[str, Any]] = None,
    priority: Optional[int] = None,
    allow_duplicate: bool = False,
) -> str:
    """
    Register a schedule for a workflow.

    This function registers both the workflow DAG and the schedule in a single
    call. When the schedule fires, the registered workflow version will be
    executed.

    Args:
        workflow_cls: The Workflow class to schedule.
        schedule_name: Unique name for this schedule. Allows multiple schedules
                       per workflow with different inputs. Must be unique within
                       a workflow.
        schedule: Either a cron expression string (e.g., "0 * * * *" for hourly)
                  or a timedelta for interval-based scheduling.
        jitter: Optional jitter window to add to each scheduled run.
        inputs: Optional keyword arguments to pass to each scheduled run.
        priority: Optional priority for queue ordering. Higher values are
                  processed first. Default is 0.
        allow_duplicate: If False (default), the scheduler skips creating a new
                         instance when one is already running for this schedule.
                         If True, always creates a new instance.

    Returns:
        The schedule ID.

    Examples:
        # Run every hour at minute 0
        await schedule_workflow(
            MyWorkflow,
            schedule_name="hourly-run",
            schedule="0 * * * *"
        )

        # Run every 5 minutes
        await schedule_workflow(
            MyWorkflow,
            schedule_name="frequent-check",
            schedule=timedelta(minutes=5)
        )

        # Multiple schedules with different inputs
        await schedule_workflow(
            MyWorkflow,
            schedule_name="small-batch",
            schedule="0 0 * * *",
            inputs={"batch_size": 100}
        )
        await schedule_workflow(
            MyWorkflow,
            schedule_name="large-batch",
            schedule="0 12 * * *",
            inputs={"batch_size": 1000}
        )

        # High priority schedule
        await schedule_workflow(
            CriticalWorkflow,
            schedule_name="critical-job",
            schedule="*/5 * * * *",
            priority=100
        )

    Raises:
        ValueError: If the cron expression is invalid, interval is non-positive,
                    or schedule_name is empty.
        ScheduleAlreadyExistsError: If a schedule with the same name already exists.
        RuntimeError: If the gRPC call fails.
    """
    if not schedule_name:
        raise ValueError("schedule_name is required")

    workflow_name = workflow_cls.short_name()

    # Build schedule definition
    schedule_def = pb2.ScheduleDefinition()
    if isinstance(schedule, str):
        schedule_def.type = pb2.SCHEDULE_TYPE_CRON
        schedule_def.cron_expression = schedule
    elif isinstance(schedule, timedelta):
        interval_seconds = int(schedule.total_seconds())
        if interval_seconds <= 0:
            raise ValueError("Interval must be positive")
        schedule_def.type = pb2.SCHEDULE_TYPE_INTERVAL
        schedule_def.interval_seconds = interval_seconds
    else:
        raise TypeError(f"schedule must be str or timedelta, got {type(schedule)}")

    if jitter is not None:
        jitter_seconds = int(jitter.total_seconds())
        if jitter_seconds < 0:
            raise ValueError("jitter must be non-negative")
        schedule_def.jitter_seconds = jitter_seconds

    # Build the workflow registration payload to ensure the DAG is registered
    # This is required for the schedule to execute - the scheduler needs a
    # registered workflow version to create instances from.
    registration = workflow_cls._build_registration_payload()

    # Build request with both registration and schedule
    request = pb2.RegisterScheduleRequest(
        workflow_name=workflow_name,
        schedule_name=schedule_name,
        schedule=schedule_def,
        registration=registration,
    )

    initial_context = workflow_cls._build_initial_context((), inputs or {})
    if initial_context.arguments:
        request.inputs.CopyFrom(initial_context)

    if priority is not None:
        request.priority = priority

    if allow_duplicate:
        request.allow_duplicate = allow_duplicate

    # Send to server
    async with ensure_singleton():
        stub = await _workflow_stub()

    try:
        response = await stub.RegisterSchedule(request, timeout=30.0)
    except aio.AioRpcError as exc:
        if exc.code() == StatusCode.ALREADY_EXISTS:
            raise ScheduleAlreadyExistsError(
                f"schedule already exists: {workflow_name}/{schedule_name}"
            ) from exc
        raise RuntimeError(f"Failed to register schedule: {exc}") from exc

    return response.schedule_id


async def pause_schedule(workflow_cls: Type[Workflow], *, schedule_name: str) -> bool:
    """
    Pause a workflow's schedule.

    The schedule will not fire until resumed. Existing running instances
    are not affected.

    Args:
        workflow_cls: The Workflow class whose schedule to pause.
        schedule_name: The name of the schedule to pause.

    Returns:
        True if a schedule was found and paused, False otherwise.

    Raises:
        ValueError: If schedule_name is empty.
        RuntimeError: If the gRPC call fails.
    """
    if not schedule_name:
        raise ValueError("schedule_name is required")

    request = pb2.UpdateScheduleStatusRequest(
        workflow_name=workflow_cls.short_name(),
        schedule_name=schedule_name,
        status=pb2.SCHEDULE_STATUS_PAUSED,
    )
    async with ensure_singleton():
        stub = await _workflow_stub()

    try:
        response = await stub.UpdateScheduleStatus(request, timeout=30.0)
    except aio.AioRpcError as exc:
        raise RuntimeError(f"Failed to pause schedule: {exc}") from exc

    return response.success


async def resume_schedule(workflow_cls: Type[Workflow], *, schedule_name: str) -> bool:
    """
    Resume a paused workflow schedule.

    Args:
        workflow_cls: The Workflow class whose schedule to resume.
        schedule_name: The name of the schedule to resume.

    Returns:
        True if a schedule was found and resumed, False otherwise.

    Raises:
        ValueError: If schedule_name is empty.
        RuntimeError: If the gRPC call fails.
    """
    if not schedule_name:
        raise ValueError("schedule_name is required")

    request = pb2.UpdateScheduleStatusRequest(
        workflow_name=workflow_cls.short_name(),
        schedule_name=schedule_name,
        status=pb2.SCHEDULE_STATUS_ACTIVE,
    )
    async with ensure_singleton():
        stub = await _workflow_stub()

    try:
        response = await stub.UpdateScheduleStatus(request, timeout=30.0)
    except aio.AioRpcError as exc:
        raise RuntimeError(f"Failed to resume schedule: {exc}") from exc

    return response.success


async def delete_schedule(workflow_cls: Type[Workflow], *, schedule_name: str) -> bool:
    """
    Delete a workflow's schedule.

    The schedule is soft-deleted and can be recreated by calling
    schedule_workflow again.

    Args:
        workflow_cls: The Workflow class whose schedule to delete.
        schedule_name: The name of the schedule to delete.

    Returns:
        True if a schedule was found and deleted, False otherwise.

    Raises:
        ValueError: If schedule_name is empty.
        RuntimeError: If the gRPC call fails.
    """
    if not schedule_name:
        raise ValueError("schedule_name is required")

    request = pb2.DeleteScheduleRequest(
        workflow_name=workflow_cls.short_name(),
        schedule_name=schedule_name,
    )
    async with ensure_singleton():
        stub = await _workflow_stub()

    try:
        response = await stub.DeleteSchedule(request, timeout=30.0)
    except aio.AioRpcError as exc:
        raise RuntimeError(f"Failed to delete schedule: {exc}") from exc

    return response.success


def _parse_iso_datetime(value: str) -> Optional[datetime]:
    """Parse an ISO 8601 datetime string, returning None if empty."""
    if not value:
        return None
    return datetime.fromisoformat(value.replace("Z", "+00:00"))


def _proto_schedule_type_to_str(
    schedule_type: "pb2.ScheduleType.V",
) -> ScheduleType:
    """Convert protobuf ScheduleType to string literal."""
    if schedule_type == pb2.SCHEDULE_TYPE_CRON:
        return "cron"
    elif schedule_type == pb2.SCHEDULE_TYPE_INTERVAL:
        return "interval"
    else:
        return "cron"  # Default fallback


def _proto_schedule_status_to_str(
    status: "pb2.ScheduleStatus.V",
) -> ScheduleStatus:
    """Convert protobuf ScheduleStatus to string literal."""
    if status == pb2.SCHEDULE_STATUS_ACTIVE:
        return "active"
    elif status == pb2.SCHEDULE_STATUS_PAUSED:
        return "paused"
    else:
        return "active"  # Default fallback


async def list_schedules(
    status_filter: Optional[ScheduleStatus] = None,
) -> List[ScheduleInfo]:
    """
    List all registered workflow schedules.

    Args:
        status_filter: Optional filter by status ("active" or "paused").
                       If None, returns all non-deleted schedules.

    Returns:
        A list of ScheduleInfo objects containing schedule details.

    Examples:
        # List all schedules
        schedules = await list_schedules()
        for s in schedules:
            print(f"{s.workflow_name}: {s.status}")

        # List only active schedules
        active = await list_schedules(status_filter="active")

        # List only paused schedules
        paused = await list_schedules(status_filter="paused")

    Raises:
        RuntimeError: If the gRPC call fails.
    """
    request = pb2.ListSchedulesRequest()
    if status_filter is not None:
        request.status_filter = status_filter

    async with ensure_singleton():
        stub = await _workflow_stub()

    try:
        response = await stub.ListSchedules(request, timeout=30.0)
    except aio.AioRpcError as exc:
        raise RuntimeError(f"Failed to list schedules: {exc}") from exc

    schedules = []
    for s in response.schedules:
        schedules.append(
            ScheduleInfo(
                id=s.id,
                workflow_name=s.workflow_name,
                schedule_name=s.schedule_name,
                schedule_type=_proto_schedule_type_to_str(s.schedule_type),
                cron_expression=s.cron_expression if s.cron_expression else None,
                interval_seconds=s.interval_seconds if s.interval_seconds else None,
                jitter_seconds=s.jitter_seconds if s.jitter_seconds else None,
                status=_proto_schedule_status_to_str(s.status),
                next_run_at=_parse_iso_datetime(s.next_run_at),
                last_run_at=_parse_iso_datetime(s.last_run_at),
                last_instance_id=s.last_instance_id if s.last_instance_id else None,
                created_at=_parse_iso_datetime(s.created_at),  # type: ignore
                updated_at=_parse_iso_datetime(s.updated_at),  # type: ignore
                allow_duplicate=s.allow_duplicate,
            )
        )

    return schedules
